# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from copy import copy

import numpy as np

# this import line is a little ridiculous...
from ..coordinate_map import (CoordinateMap, AffineTransform, compose, product,
                              append_io_dim, drop_io_dim, equivalent,
                              shifted_domain_origin, shifted_range_origin,
                              CoordMapMaker, CoordMapMakerError,
                              _as_coordinate_map, AxisError, _fix0,
                              axmap, orth_axes, input_axis_index, io_axis_indices)

from ..coordinate_system import (CoordinateSystem, CoordinateSystemError,
                                 CoordSysMaker, CoordSysMakerError)

# shortcut
CS = CoordinateSystem

from nose.tools import (assert_true, assert_equal, assert_raises,
                        assert_false)

from numpy.testing import (assert_array_equal, assert_almost_equal, dec)


class empty(object):
    pass

# object to hold module global setup
E = empty()


def setup():
    def f(x):
        return 2*x
    def g(x):
        return x/2.0
    x = CoordinateSystem('x', 'x')
    E.a = CoordinateMap(x, x, f)
    E.b = CoordinateMap(x, x, f, inverse_function=g)
    E.c = CoordinateMap(x, x, g)
    E.d = CoordinateMap(x, x, g, inverse_function=f)
    E.e = AffineTransform.identity('ijk')
    A = np.identity(4)
    A[0:3] = np.random.standard_normal((3,4))
    E.mapping = AffineTransform.from_params('ijk' ,'xyz', A)
    E.singular = AffineTransform.from_params('ijk', 'xyzt',
                                    np.array([[ 0,  1,  2,  3],
                                              [ 4,  5,  6,  7],
                                              [ 8,  9, 10, 11],
                                              [ 8,  9, 10, 11],
                                              [ 0,  0,  0,  1]]))


def test_shift_origin():
    CS = CoordinateSystem

    A = np.random.standard_normal((5,6))
    A[-1] = [0,0,0,0,0,1]

    aff1 = AffineTransform(CS('ijklm', 'oldorigin'), CS('xyzt'), A)
    difference = np.random.standard_normal(5)
    point_in_old_basis = np.random.standard_normal(5)

    for aff in [aff1, _as_coordinate_map(aff1)]:
        # The same affine transformation with a different origin for its domain
        shifted_aff = shifted_domain_origin(aff, difference, 'neworigin')
        # This is the relationship between coordinates in old and new origins
        assert_almost_equal(shifted_aff(point_in_old_basis),
                            aff(point_in_old_basis+difference))
        assert_almost_equal(shifted_aff(point_in_old_basis-difference),
                            aff(point_in_old_basis))
    # OK, now for the range
    A = np.random.standard_normal((5,6))
    A[-1] = [0,0,0,0,0,1]
    aff2 = AffineTransform(CS('ijklm', 'oldorigin'), CS('xyzt'), A)
    difference = np.random.standard_normal(4)
    for aff in [aff2, _as_coordinate_map(aff2)]:
    # The same affine transformation with a different origin for its domain
        shifted_aff = shifted_range_origin(aff, difference, 'neworigin')
        # Let's check that things work
        point_in_old_basis = np.random.standard_normal(5)
        # This is the relation ship between coordinates in old and new origins
        assert_almost_equal(shifted_aff(point_in_old_basis),
                            aff(point_in_old_basis)-difference)
        assert_almost_equal(shifted_aff(point_in_old_basis)+difference,
                            aff(point_in_old_basis))


def test_renamed():
    # Renaming domain and range
    A = AffineTransform.from_params('ijk', 'xyz', np.identity(4))
    ijk = CoordinateSystem('ijk')
    xyz = CoordinateSystem('xyz')
    C = CoordinateMap(ijk, xyz, np.log)
    for B in [A,C]:
        B_re = B.renamed_domain({'i':'foo'})
        assert_equal(B_re.function_domain.coord_names, ('foo', 'j', 'k'))
        B_re = B.renamed_domain({'i':'foo','j':'bar'})
        assert_equal(B_re.function_domain.coord_names, ('foo', 'bar', 'k'))
        B_re = B.renamed_range({'y':'foo'})
        assert_equal(B_re.function_range.coord_names, ('x', 'foo', 'z'))
        B_re = B.renamed_range({0:'foo',1:'bar'})
        assert_equal(B_re.function_range.coord_names, ('foo', 'bar', 'z'))
        B_re = B.renamed_domain({0:'foo',1:'bar'})
        assert_equal(B_re.function_domain.coord_names, ('foo', 'bar', 'k'))
        B_re = B.renamed_range({'y':'foo','x':'bar'})
        assert_equal(B_re.function_range.coord_names, ('bar', 'foo', 'z'))
        assert_raises(ValueError, B.renamed_range, {'foo':'y'})
        assert_raises(ValueError, B.renamed_domain, {'foo':'y'})


def test_calling_shapes():
    cs2d = CS('ij')
    cs1d = CS('i')
    cm2d = CoordinateMap(cs2d, cs2d, lambda x : x+1)
    cm1d2d = CoordinateMap(cs1d, cs2d,
                           lambda x : np.concatenate((x, x), axis=-1))
    at2d = AffineTransform(cs2d, cs2d, np.array([[1, 0, 1],
                                                 [0, 1, 1],
                                                 [0, 0, 1]]))
    at1d2d = AffineTransform(cs1d, cs2d, np.array([[1,0],
                                                   [0,1],
                                                   [0,1]]))
    # test coordinate maps and affine transforms
    for xfm2d, xfm1d2d in ((cm2d, cm1d2d), (at2d, at1d2d)):
        arr = np.array([0, 1])
        assert_array_equal(xfm2d(arr), [1, 2])
        # test lists work too
        res = xfm2d([0, 1])
        assert_array_equal(res, [1, 2])
        # and return arrays (by checking shape attribute)
        assert_equal(res.shape, (2,))
        # maintaining input shape
        arr_long = arr[None, None, :]
        assert_array_equal(xfm2d(arr_long), arr_long + 1)
        # wrong shape array raises error
        assert_raises(CoordinateSystemError, xfm2d, np.zeros((3,)))
        assert_raises(CoordinateSystemError, xfm2d, np.zeros((3,3)))
        # 1d to 2d
        arr = np.array(1)
        assert_array_equal(xfm1d2d(arr), [1,1] )
        arr_long = arr[None, None, None]
        assert_array_equal(xfm1d2d(arr_long), np.ones((1,1,2)))
        # wrong shape array raises error.  Note 1d input requires size 1
        # as final axis
        assert_raises(CoordinateSystemError, xfm1d2d, np.zeros((3,)))
        assert_raises(CoordinateSystemError, xfm1d2d, np.zeros((3,2)))


def test_call():
    value = 10
    assert_true(np.allclose(E.a(value), 2*value))
    assert_true(np.allclose(E.b(value), 2*value))
    # FIXME: this shape just below is not
    # really expected for a CoordinateMap
    assert_true(np.allclose(E.b([value]), 2*value))
    assert_true(np.allclose(E.c(value), value/2))
    assert_true(np.allclose(E.d(value), value/2))
    value = np.array([1., 2., 3.])
    assert_true(np.allclose(E.e(value), value))
    # check that error raised for wrong shape
    value = np.array([1., 2.,])
    assert_raises(CoordinateSystemError, E.e, value)


def test_compose():
    value = np.array([[1., 2., 3.]]).T
    aa = compose(E.a, E.a)
    assert_true(aa.inverse() is None)
    assert_almost_equal(aa(value), 4*value)
    ab = compose(E.a,E.b)
    assert_true(ab.inverse() is None)
    assert_almost_equal(ab(value), 4*value)
    ac = compose(E.a,E.c)
    assert_true(ac.inverse() is None)
    assert_almost_equal(ac(value), value)
    bb = compose(E.b,E.b)
    #    yield assert_true, bb.inverse() is not None
    aff1 = np.diag([1,2,3,1])
    affine1 = AffineTransform.from_params('ijk', 'xyz', aff1)
    aff2 = np.diag([4,5,6,1])
    affine2 = AffineTransform.from_params('xyz', 'abc', aff2)
    # compose mapping from 'ijk' to 'abc'
    compcm = compose(affine2, affine1)
    assert_equal(compcm.function_domain.coord_names, ('i', 'j', 'k'))
    assert_equal(compcm.function_range.coord_names, ('a', 'b', 'c'))
    assert_almost_equal(compcm.affine, np.dot(aff2, aff1))
    # check invalid coordinate mappings
    assert_raises(ValueError, compose, affine1, affine2)
    assert_raises(ValueError, compose, affine1, 'foo')
    cm1 = CoordinateMap(CoordinateSystem('ijk'),
                        CoordinateSystem('xyz'), np.log)
    cm2 = CoordinateMap(CoordinateSystem('xyz'),
                        CoordinateSystem('abc'), np.exp)
    assert_raises(ValueError, compose, cm1, cm2)


def test__eq__():
    yield assert_true, E.a == E.a
    yield assert_false, E.a != E.a

    yield assert_false, E.a == E.b
    yield assert_true, E.a != E.b

    yield assert_true, E.singular == E.singular
    yield assert_false, E.singular != E.singular

    A = AffineTransform.from_params('ijk', 'xyz', np.diag([4,3,2,1]))
    B = AffineTransform.from_params('ijk', 'xyz', np.diag([4,3,2,1]))

    yield assert_true, A == B
    yield assert_false, A != B


def test_similar_to():
    in_cs = CoordinateSystem('ijk', 'in', np.float32)
    in_cs2 = CoordinateSystem('ijk', 'another name', np.float32)
    out_cs = CoordinateSystem('xyz', 'out', np.float32)
    out_cs2 = CoordinateSystem('xyz', 'again another', np.float32)
    for klass, arg0, arg1 in ((CoordinateMap,
                               lambda x : x + 1, lambda x : x + 2),
                             (AffineTransform,
                              np.eye(4), np.diag([1, 2, 3, 1]))):
        c0 = klass(in_cs, out_cs, arg0)
        c1 = klass(in_cs, out_cs, arg0)
        assert_true(c0.similar_to(c1))
        c1b = klass(in_cs, out_cs, arg1)
        assert_false(c0.similar_to(c1b))
        c2 = klass(in_cs2, out_cs, arg0)
        assert_true(c0.similar_to(c2))
        c3 = klass(in_cs, out_cs2, arg0)
        assert_true(c0.similar_to(c3))


def test_isinvertible():
    yield assert_false, E.a.inverse()
    yield assert_true, E.b.inverse()
    yield assert_false, E.c.inverse()
    yield assert_true, E.d.inverse()
    yield assert_true, E.e.inverse()
    yield assert_true, E.mapping.inverse()
    yield assert_false, E.singular.inverse()


def test_inverse1():
    inv = lambda a: a.inverse()
    yield assert_true, inv(E.a) is None
    yield assert_true, inv(E.c) is None
    inv_b = E.b.inverse()
    inv_d = E.d.inverse()
    ident_b = compose(inv_b,E.b)
    ident_d = compose(inv_d,E.d)
    value = np.array([[1., 2., 3.]]).T
    yield assert_true, np.allclose(ident_b(value), value)
    yield assert_true, np.allclose(ident_d(value), value)


def test_compose_cmap():
    value = np.array([1., 2., 3.])
    b = compose(E.e, E.e)
    assert_true(np.allclose(b(value), value))


def test_inverse2():
    assert_true(np.allclose(E.e.affine, E.e.inverse().inverse().affine))


def voxel_to_world():
    # utility function for generating trivial CoordinateMap
    incs = CoordinateSystem('ijk', 'voxels')
    outcs = CoordinateSystem('xyz', 'world')
    map = lambda x: x + 1
    inv = lambda x: x - 1
    return incs, outcs, map, inv


def test_comap_init():
    # Test mapping and non-mapping functions
    incs, outcs, map, inv = voxel_to_world()
    cm = CoordinateMap(incs, outcs, map, inv)
    yield assert_equal, cm.function, map
    yield assert_equal, cm.function_domain, incs
    yield assert_equal, cm.function_range, outcs
    yield assert_equal, cm.inverse_function, inv
    yield assert_raises, ValueError, CoordinateMap, incs, outcs, 'foo', inv
    yield assert_raises, ValueError, CoordinateMap, incs, outcs, map, 'bar'


def test_comap_cosys():
    # Check we can pass in coordinate names instead of coordinate systems
    d_sys = CoordinateSystem('ijk')
    r_sys = CoordinateSystem('xyz')
    fn = lambda x : x+1
    cm = CoordinateMap(d_sys, r_sys, fn)
    assert_equal(CoordinateMap('ijk', 'xyz', fn), cm)
    assert_equal(CoordinateMap(d_sys, 'xyz', fn), cm)
    assert_equal(CoordinateMap('ijk', r_sys, fn), cm)
    aff = np.diag([2,3,4,1])
    cm = AffineTransform(d_sys, r_sys, aff)
    assert_equal(AffineTransform('ijk', 'xyz', aff), cm)
    assert_equal(AffineTransform(d_sys, 'xyz', aff), cm)
    assert_equal(AffineTransform('ijk', r_sys, aff), cm)


def test_comap_copy():
    import copy
    incs, outcs, map, inv = voxel_to_world()
    cm = CoordinateMap(incs, outcs, inv, map)
    cmcp = copy.copy(cm)
    yield assert_equal, cmcp.function, cm.function
    yield assert_equal, cmcp.function_domain, cm.function_domain
    yield assert_equal, cmcp.function_range, cm.function_range
    yield assert_equal, cmcp.inverse_function, cm.inverse_function


#
# AffineTransform tests
#

def affine_v2w():
    # utility function
    incs = CoordinateSystem('ijk', 'voxels')
    outcs = CoordinateSystem('xyz', 'world')
    aff = np.diag([1, 2, 4, 1])
    aff[:3, 3] = [11, 12, 13]
    """array([[ 1,  0,  0, 11],
       [ 0,  2,  0, 12],
       [ 0,  0,  4, 13],
       [ 0,  0,  0,  1]])
    """
    return incs, outcs, aff


def test_affine_init():
    incs, outcs, aff = affine_v2w()
    cm = AffineTransform(incs, outcs, aff)
    assert_equal(cm.function_domain, incs)
    assert_equal(cm.function_range, outcs)
    assert_array_equal(cm.affine, aff)
    badaff = np.diag([1,2])
    assert_raises(ValueError, AffineTransform, incs, outcs, badaff)


def test_affine_bottom_row():
    # homogeneous transformations have bottom rows all zero
    # except the last one
    assert_raises(ValueError, AffineTransform.from_params,
                  'ij',  'x', np.array([[3,4,5],[1,1,1]]))


def test_affine_inverse():
    incs, outcs, aff = affine_v2w()
    inv = np.linalg.inv(aff)
    cm = AffineTransform(incs, outcs, aff)
    x = np.array([10, 20, 30], np.float)
    x_roundtrip = cm(cm.inverse()(x))
    assert_almost_equal(x_roundtrip, x)
    badaff = np.array([[1,2,3],[0,0,1]])
    badcm = AffineTransform(CoordinateSystem('ij'),
                            CoordinateSystem('x'),
                            badaff)
    assert_equal(badcm.inverse(), None)


def test_affine_from_params():
    incs, outcs, aff = affine_v2w()
    cm = AffineTransform.from_params('ijk', 'xyz', aff)
    assert_array_equal(cm.affine, aff)
    badaff = np.array([[1,2,3],[4,5,6]])
    assert_raises(ValueError,
                  AffineTransform.from_params, 'ijk', 'xyz', badaff)


def test_affine_start_step():
    incs, outcs, aff = affine_v2w()
    start = aff[:3, 3]
    step = aff.diagonal()[:3]
    cm = AffineTransform.from_start_step(incs.coord_names, outcs.coord_names,
                                start, step)
    assert_array_equal(cm.affine, aff)
    assert_raises(ValueError, AffineTransform.from_start_step, 'ijk', 'xy',
                  start, step)


def test_affine_identity():
    aff = AffineTransform.identity('ijk')
    assert_array_equal(aff.affine, np.eye(4))
    assert_equal(aff.function_domain, aff.function_range)
    # AffineTransform's aren't CoordinateMaps, so
    # they don't have "function" attributes
    assert_false(hasattr(aff, 'function'))


def test_affine_copy():
    incs, outcs, aff = affine_v2w()
    cm = AffineTransform(incs, outcs, aff)
    import copy
    cmcp = copy.copy(cm)
    assert_array_equal(cmcp.affine, cm.affine)
    assert_equal(cmcp.function_domain, cm.function_domain)
    assert_equal(cmcp.function_range, cm.function_range)


#
# Module level functions
#

def test_reordered_domain():
    incs, outcs, map, inv = voxel_to_world()
    cm = CoordinateMap(incs, outcs, map, inv)
    recm = cm.reordered_domain('jki')
    yield assert_equal, recm.function_domain.coord_names, ('j', 'k', 'i')
    yield assert_equal, recm.function_range.coord_names, outcs.coord_names
    yield assert_equal, recm.function_domain.name, incs.name
    yield assert_equal, recm.function_range.name, outcs.name
    # default reverse reorder
    recm = cm.reordered_domain()
    yield assert_equal, recm.function_domain.coord_names, ('k', 'j', 'i')
    # reorder with order as indices
    recm = cm.reordered_domain([2,0,1])
    yield assert_equal, recm.function_domain.coord_names, ('k', 'i', 'j')


def test_str():
    result = """AffineTransform(
   function_domain=CoordinateSystem(coord_names=('i', 'j', 'k'), name='', coord_dtype=float64),
   function_range=CoordinateSystem(coord_names=('x', 'y', 'z'), name='', coord_dtype=float64),
   affine=array([[ 1.,  0.,  0.,  0.],
                 [ 0.,  1.,  0.,  0.],
                 [ 0.,  0.,  1.,  0.],
                 [ 0.,  0.,  0.,  1.]])
)"""
    domain = CoordinateSystem('ijk')
    range = CoordinateSystem('xyz')
    affine = np.identity(4)
    affine_mapping = AffineTransform(domain, range, affine)
    assert_equal(result, str(affine_mapping))

    cmap = CoordinateMap(domain, range, np.exp, np.log)
    result="""CoordinateMap(
   function_domain=CoordinateSystem(coord_names=('i', 'j', 'k'), name='', coord_dtype=float64),
   function_range=CoordinateSystem(coord_names=('x', 'y', 'z'), name='', coord_dtype=float64),
   function=<ufunc 'exp'>,
   inverse_function=<ufunc 'log'>
  )"""
    cmap = CoordinateMap(domain, range, np.exp)
    result="""CoordinateMap(
   function_domain=CoordinateSystem(coord_names=('i', 'j', 'k'), name='', coord_dtype=float64),
   function_range=CoordinateSystem(coord_names=('x', 'y', 'z'), name='', coord_dtype=float64),
   function=<ufunc 'exp'>
  )"""
    assert_equal(result, repr(cmap))


def test_reordered_range():
    incs, outcs, map, inv = voxel_to_world()
    cm = CoordinateMap(incs, outcs, inv, map)
    recm = cm.reordered_range('yzx')
    yield assert_equal, recm.function_domain.coord_names, incs.coord_names
    yield assert_equal, recm.function_range.coord_names, ('y', 'z', 'x')
    yield assert_equal, recm.function_domain.name, incs.name
    yield assert_equal, recm.function_range.name, outcs.name
    # default reverse order
    recm = cm.reordered_range()
    yield assert_equal, recm.function_range.coord_names, ('z', 'y', 'x')
    # reorder with indicies
    recm = cm.reordered_range([2,0,1])
    yield assert_equal, recm.function_range.coord_names, ('z', 'x', 'y')


def test_product():
    affine1 = AffineTransform.from_params('i', 'x', np.diag([2, 1]))
    affine2 = AffineTransform.from_params('j', 'y', np.diag([3, 1]))
    affine = product(affine1, affine2)
    cm1 = CoordinateMap(CoordinateSystem('i'),
                        CoordinateSystem('x'),
                        np.log)
    cm2 = CoordinateMap(CoordinateSystem('j'),
                        CoordinateSystem('y'),
                        np.log)
    cm = product(cm1, cm2)
    assert_equal(affine.function_domain.coord_names, ('i', 'j'))
    assert_equal(affine.function_range.coord_names, ('x', 'y'))
    assert_almost_equal(cm([3,4]), np.log([3,4]))
    assert_almost_equal(cm.function([[3,4],[5,6]]), np.log([[3,4],[5,6]]))
    assert_equal(affine.function_domain.coord_names, ('i', 'j'))
    assert_equal(affine.function_range.coord_names, ('x', 'y'))
    assert_array_equal(affine.affine, np.diag([2, 3, 1]))
    # Test name argument
    for m1, m2 in ((affine1, affine2), (cm1, cm2), (affine1, cm2)):
        cm = product(m1, m2)
        assert_equal(cm.function_domain.name, 'product')
        assert_equal(cm.function_range.name, 'product')
        cm = product(m1, m2, input_name='name0')
        assert_equal(cm.function_domain.name, 'name0')
        assert_equal(cm.function_range.name, 'product')
        cm = product(m1, m2, output_name='name1')
        assert_equal(cm.function_domain.name, 'product')
        assert_equal(cm.function_range.name, 'name1')
        assert_raises(TypeError, product, m1, m2, whatgains='name0')


def test_equivalent():
    ijk = CoordinateSystem('ijk')
    xyz = CoordinateSystem('xyz')
    T = np.random.standard_normal((4,4))
    T[-1] = [0,0,0,1]
    A = AffineTransform(ijk, xyz, T)

    # now, cycle through
    # all possible permutations of
    # 'ijk' and 'xyz' and confirm that
    # the mapping is equivalent

    yield assert_false, equivalent(A, A.renamed_domain({'i':'foo'}))

    try:
        import itertools
        for pijk in itertools.permutations('ijk'):
            for pxyz in itertools.permutations('xyz'):
                B = A.reordered_domain(pijk).reordered_range(pxyz)
                yield assert_true, equivalent(A, B)
    except (ImportError, AttributeError):
        # just do some if we can't find itertools, or if itertools
        # doesn't have permutations
        for pijk in ['ikj', 'kij']:
            for pxyz in ['xzy', 'yxz']:
                B = A.reordered_domain(pijk).reordered_range(pxyz)
                yield assert_true, equivalent(A, B)


def test_as_coordinate_map():

    ijk = CoordinateSystem('ijk')
    xyz = CoordinateSystem('xyz')

    A = np.random.standard_normal((4,4))

    # bottom row of A is not [0,0,0,1]
    yield assert_raises, ValueError, AffineTransform, ijk, xyz, A

    A[-1] = [0,0,0,1]

    aff = AffineTransform(ijk, xyz, A)
    _cmapA = _as_coordinate_map(aff)
    yield assert_true, isinstance(_cmapA, CoordinateMap)
    yield assert_true, _cmapA.inverse_function != None

    # a non-invertible one

    B = A[1:]
    xy = CoordinateSystem('xy')
    affB = AffineTransform(ijk, xy, B)
    _cmapB = _as_coordinate_map(affB)

    yield assert_true, isinstance(_cmapB, CoordinateMap)
    yield assert_true, _cmapB.inverse_function == None


def test_cm__setattr__raise_error():
    # CoordinateMap has all read-only attributes

    # AffineTransform has some properties and it seems
    # the same __setattr__ doesn't work for it.
    ijk = CoordinateSystem('ijk')
    xyz = CoordinateSystem('xyz')

    cm = CoordinateMap(ijk, xyz, np.exp)

    yield assert_raises, AttributeError, cm.__setattr__, "function_range", xyz


def test_append_io_dim():
    aff = np.diag([1,2,3,1])
    in_dims = tuple('ijk')
    out_dims = tuple('xyz')
    cm = AffineTransform.from_params(in_dims, out_dims, aff)
    cm2 = append_io_dim(cm, 'l', 't')
    assert_array_equal(cm2.affine, np.diag([1,2,3,1,1]))
    assert_equal(cm2.function_range.coord_names, out_dims + ('t',))
    assert_equal(cm2.function_domain.coord_names, in_dims + ('l',))
    cm2 = append_io_dim(cm, 'l', 't', 9, 5)
    a2 = np.diag([1,2,3,5,1])
    a2[3,4] = 9
    assert_array_equal(cm2.affine, a2)
    assert_equal(cm2.function_range.coord_names, out_dims + ('t',))
    assert_equal(cm2.function_domain.coord_names, in_dims + ('l',))
    # non square case
    aff = np.array([[2,0,0],
                    [0,3,0],
                    [0,0,1],
                    [0,0,1]])
    cm = AffineTransform.from_params('ij', 'xyz', aff)
    cm2 = append_io_dim(cm, 'q', 't', 9, 5)
    a2 = np.array([[2,0,0,0],
                   [0,3,0,0],
                   [0,0,0,1],
                   [0,0,5,9],
                   [0,0,0,1]])
    assert_array_equal(cm2.affine, a2)
    assert_equal(cm2.function_range.coord_names, tuple('xyzt'))
    assert_equal(cm2.function_domain.coord_names, tuple('ijq'))


def test__fix0():
    # Test routine to fix possible zero TR in affine
    assert_array_equal(_fix0(np.diag([1, 2, 3, 1])), np.diag([1, 2, 3, 1]))
    assert_array_equal(_fix0(np.diag([0, 2, 3, 1])), np.diag([1, 2, 3, 1]))
    assert_array_equal(_fix0(np.diag([1, 0, 3, 1])), np.diag([1, 1, 3, 1]))
    assert_array_equal(_fix0(np.diag([1, 2, 0, 1])), np.diag([1, 2, 1, 1]))
    aff = [[1, 0, 0, 10],
           [0, 0, 0, 11],
           [0, 0, 0, 1]]
    assert_array_equal(_fix0(aff), aff)
    aff = [[1, 0, 0, 10],
           [0, 2, 0, 11],
           [0, 0, 0, 12],
           [0, 0, 0, 1]]
    assert_array_equal(_fix0(aff),
                       [[1, 0, 0, 10],
                        [0, 2, 0, 11],
                        [0, 0, 1, 12],
                        [0, 0, 0, 1]])
    eps = np.finfo(np.float64).eps
    aff[2][2] = eps
    assert_array_equal(_fix0(aff), aff)


def test_drop_io_dim():
    # test ordinary case of 4d to 3d
    cm4d = AffineTransform.from_params('ijkl', 'xyzt', np.diag([1,2,3,4,1]))
    cm3d = drop_io_dim(cm4d, 't')
    assert_array_equal(cm3d.affine, np.diag([1, 2, 3, 1]))
    cm3d = drop_io_dim(cm4d, 'l')
    assert_array_equal(cm3d.affine, np.diag([1, 2, 3, 1]))
    cm3d = drop_io_dim(cm4d, 3)
    assert_array_equal(cm3d.affine, np.diag([1, 2, 3, 1]))
    cm3d = drop_io_dim(cm4d, -1)
    assert_array_equal(cm3d.affine, np.diag([1, 2, 3, 1]))
    # 3d to 2d
    cm3d = AffineTransform.from_params('ijk', 'xyz', np.diag([1,2,3,1]))
    cm2d = drop_io_dim(cm3d, 'z')
    assert_array_equal(cm2d.affine, np.diag([1, 2, 1]))
    # test zero scaling for dropped dimension
    cm3d = AffineTransform.from_params('ijk', 'xyz', np.diag([1, 2, 0, 1]))
    cm2d = drop_io_dim(cm3d, 'z')
    assert_array_equal(cm2d.affine, np.diag([1, 2, 1]))
    # test not diagonal but orthogonal
    aff = np.array([[1, 0, 0, 0],
                    [0, 0, 2, 0],
                    [0, 3, 0, 0],
                    [0, 0, 0, 1]])
    cm3d = AffineTransform.from_params('ijk', 'xyz', aff)
    cm2d = drop_io_dim(cm3d, 'z')
    assert_array_equal(cm2d.affine, np.diag([1, 2, 1]))
    cm2d = drop_io_dim(cm3d, 'k')
    assert_array_equal(cm2d.affine, np.diag([1, 3, 1]))
    # and with zeros scaling fix for orthogonal dropped dimension
    aff[2] = 0
    cm3d = AffineTransform.from_params('ijk', 'xyz', aff)
    cm2d = drop_io_dim(cm3d, 'z')
    assert_array_equal(cm2d.affine, np.diag([1, 2, 1]))
    # Unless told otherwise
    cm2d = drop_io_dim(cm3d, 'z', fix0=False)
    # In this case we drop z because it has no matching input
    assert_array_equal(cm2d.affine, [[1, 0, 0, 0],
                                     [0, 0, 2, 0],
                                     [0, 0, 0, 1]])
    # Don't zero-fix untested dimensions
    cm2d = drop_io_dim(cm3d, 'y', fix0=True)
    assert_array_equal(cm2d.affine, np.diag([1, 0, 1]))
    # Test test for ambiguous coordinate names
    # This one is OK because they match
    cm3d = AffineTransform.from_params('ijk', 'iyz', np.diag([1, 2, 3, 1]))
    cm2d = drop_io_dim(cm3d, 'i')
    assert_array_equal(cm2d.affine, np.diag([2, 3, 1]))
    # Here they don't match and this raises an error
    cm3d = AffineTransform.from_params('ijk', 'xiz', np.diag([1, 2, 3, 1]))
    assert_raises(AxisError, drop_io_dim, cm3d, 'i')
    # Dropping input or outputs that have no matching dimensions is also OK
    aff = np.array([[1, .1, 0, 10],
                    [.1, 0, 0, 11],
                    [ 0, 3, 0, 12],
                    [ 0, 0, 0, 1]])
    cm3d = AffineTransform.from_params('ijk', 'xyz', aff)
    cm2d = drop_io_dim(cm3d, 'k')
    assert_array_equal(cm2d.affine, [[1, .1, 10],
                                     [.1, 0, 11],
                                     [ 0, 3, 12],
                                     [ 0, 0, 1]])
    aff = np.array([[1, .1, 0, 10],
                    [0, 0, 0, 11],
                    [0, 3, .1, 12],
                    [0, 0, 0, 1]])
    cm3d = AffineTransform.from_params('ijk', 'xyz', aff)
    cm2d = drop_io_dim(cm3d, 'y')
    assert_array_equal(cm2d.affine, [[1, .1, 0, 10],
                                     [ 0, 3, .1, 12],
                                     [ 0, 0, 0, 1]])


def test_axmap():
    # Test mapping between axes
    cmap = AffineTransform('ijk', 'xyz', np.eye(4))
    assert_equal(axmap(cmap), {0: 0, 1:1, 2:2,
                               'i': 0, 'j': 1, 'k': 2})
    assert_equal(axmap(cmap, 'out2in'), {0: 0, 1:1, 2:2,
                                         'x': 0, 'y': 1, 'z': 2})
    assert_equal(axmap(cmap, 'both'), ({0: 0, 1:1, 2:2,
                                        'i': 0, 'j': 1, 'k': 2},
                                       {0: 0, 1:1, 2:2,
                                        'x': 0, 'y': 1, 'z': 2}))
    cmap = AffineTransform('ijk', 'xyz', [[0, 1, 0, 0],
                                          [0, 0, 1, 0],
                                          [1, 0, 0, 0],
                                          [0, 0, 0, 1]])
    assert_equal(axmap(cmap), {0: 2, 1: 0, 2: 1,
                               'i': 2, 'j': 0, 'k': 1})
    assert_equal(axmap(cmap, 'out2in'), {2: 0, 0: 1, 1: 2,
                                         'z': 0, 'x': 1, 'y': 2})
    # Test in presence of nasty zero
    cmap = AffineTransform('ijk', 'xyz', np.diag([2, 3, 0, 1]))
    # Default is to fix zero
    assert_equal(axmap(cmap), {0: 0, 1: 1, 2: 2,
                               'i': 0, 'j': 1, 'k': 2})
    assert_equal(axmap(cmap, fix0=True), {0: 0, 1: 1, 2: 2,
                                          'i': 0, 'j': 1, 'k': 2})
    assert_equal(axmap(cmap, 'out2in'), {0: 0, 1: 1, 2: 2,
                                         'x': 0, 'y': 1, 'z': 2})
    # If turned off, we can't find the axis anymore
    assert_equal(axmap(cmap, fix0=False), {0: 0, 1: 1, 2: None,
                                           'i': 0, 'j': 1, 'k': None})
    assert_equal(axmap(cmap, 'out2in', fix0=False), {0: 0, 1: 1, 2: None,
                                                    'x': 0, 'y': 1, 'z': None})
    # Need in2out or out2in as action strings
    assert_raises(ValueError, axmap, cmap, 'do what exactly?')
    # Non-square
    cmap = AffineTransform('ij', 'xyz', [[0, 1, 0],
                                         [0, 0, 0],
                                         [1, 0, 0],
                                         [0, 0, 1]])
    assert_equal(axmap(cmap), {0: 2, 1: 0,
                               'i': 2, 'j': 0})
    assert_equal(axmap(cmap, 'out2in'), {0: 1, 1: None, 2: 0,
                                         'x': 1, 'y': None, 'z': 0})
    cmap = AffineTransform('ijk', 'xy', [[0, 1, 0, 0],
                                         [0, 0, 1, 0],
                                         [0, 0, 0, 1]])
    assert_equal(axmap(cmap), {0: None, 1: 0, 2: 1,
                               'i': None, 'j': 0, 'k': 1})
    assert_equal(axmap(cmap, 'out2in'), {0: 1, 1: 2,
                                         'x': 1, 'y': 2})
    # What happens if there are ties?
    cmap = AffineTransform('ijk', 'xyz', [[0, 1, 0, 0],
                                          [0, 1, 0, 0],
                                          [1, 0, 0, 0],
                                          [0, 0, 0, 1]])
    assert_equal(axmap(cmap), {0: 2, 1: 0, 2: None,
                               'i': 2, 'j': 0, 'k': None})
    assert_equal(axmap(cmap, 'out2in'), {0: 1, 1: None, 2: 0,
                                         'x': 1, 'y': None, 'z': 0})


def test_orth_axes():
    # Test for test of orthogality of in, out axis to rest of affine
    # Check 3,3, 2, 3, and that negative values don't confuse
    for aff in (np.eye(4), np.diag([2, 3, 1]), np.eye(4) * -1):
        for i in range(aff.shape[0]-1):
            assert_true(orth_axes(i, i, aff))
    assert_true(orth_axes(2, 2, np.diag([2, 3, 0, 1])))
    assert_false(orth_axes(2, 2, np.diag([2, 3, 0, 1]), False))
    aff = np.eye(4)
    assert_true(orth_axes(0, 0, aff))
    aff[0, 1] = 1e-4
    assert_false(orth_axes(0, 0, aff))
    assert_true(orth_axes(0, 0, aff, tol=2e-4))
    aff[1, 0] = 3e-4
    assert_false(orth_axes(0, 0, aff))


def test_input_axis_index():
    # Test routine to map name to input axis
    cmap = AffineTransform('ijk', 'xyz', np.eye(4))
    for i, in_name, out_name in zip(range(3), 'ijk', 'xyz'):
        assert_equal(input_axis_index(cmap, in_name), i)
        assert_equal(input_axis_index(cmap, out_name), i)
    flipped = [[0, 0, 1, 1], [0, 1, 0, 2], [1, 0, 0, 3], [0, 0, 0, 1]]
    cmap_f = AffineTransform('ijk', 'xyz', flipped)
    for i, in_name, out_name in zip(range(3), 'ijk', 'zyx'):
        assert_equal(input_axis_index(cmap_f, in_name), i)
        assert_equal(input_axis_index(cmap_f, out_name), i)
    # Names can be same in input and output but they must match
    cmap_m = AffineTransform('ijk', 'kji', flipped)
    for i, in_name, out_name in zip(range(3), 'ijk', 'ijk'):
        assert_equal(input_axis_index(cmap_m, in_name), i)
        assert_equal(input_axis_index(cmap_m, out_name), i)
    # If they don't match, AxisError
    cmap_b = AffineTransform('ijk', 'xiz', np.eye(4))
    assert_equal(input_axis_index(cmap_m, 'j'), 1)
    assert_raises(AxisError, input_axis_index, cmap_b, 'i')
    # Name not found, AxisError
    assert_raises(AxisError, input_axis_index, cmap_b, 'q')
    # 0 leads to no match if fix0 turned off
    cmap_z = AffineTransform('ijk', 'xyz', np.diag([2, 3, 0, 1]))
    assert_equal(input_axis_index(cmap_z, 'z'), 2)
    assert_equal(input_axis_index(cmap_z, 'z', fix0=True), 2)
    assert_raises(AxisError, input_axis_index, cmap_z, 'z', fix0=False)
    # Other axes not affected in presence of 0
    assert_equal(input_axis_index(cmap_z, 'y'), 1)


def test_io_axis_indices():
    # Test routine to get input and output axis indices
    cmap = AffineTransform('ijk', 'xyz', np.eye(4))
    for i, in_name, out_name in zip(range(3), 'ijk', 'xyz'):
        assert_equal(io_axis_indices(cmap, i), (i, i))
        assert_equal(io_axis_indices(cmap, in_name), (i, i))
        assert_equal(io_axis_indices(cmap, out_name), (i, i))
    flipped = [[0, 0, 1, 1], [0, 1, 0, 2], [1, 0, 0, 3], [0, 0, 0, 1]]
    cmap_f = AffineTransform('ijk', 'xyz', flipped)
    for i, in_name, out_name in zip(range(3), 'ijk', 'xyz'):
        assert_equal(io_axis_indices(cmap_f, i), (i, 2-i))
        assert_equal(io_axis_indices(cmap_f, in_name), (i, 2-i))
        assert_equal(io_axis_indices(cmap_f, out_name), (2-i, i))
    # Names can be same in input and output but they must match
    cmap_m = AffineTransform('ijk', 'kji', flipped)
    for i, in_name, out_name in zip(range(3), 'ijk', 'kji'):
        assert_equal(io_axis_indices(cmap_m, i), (i, 2-i))
        assert_equal(io_axis_indices(cmap_m, in_name), (i, 2-i))
        assert_equal(io_axis_indices(cmap_m, out_name), (2-i, i))
    # If they don't match, AxisError, if selecting by name
    cmap_b = AffineTransform('ijk', 'xiz', np.eye(4))
    assert_raises(AxisError, io_axis_indices, cmap_b, 'i')
    # ... but not if name corresponds
    assert_equal(io_axis_indices(cmap_b, 'k'), (2, 2))
    # ... or if input name not found in output
    assert_equal(io_axis_indices(cmap_b, 'j'), (1, 1))
    # ... or if selecting by number
    assert_equal(io_axis_indices(cmap_b, 0), (0, 0))
    # Name not found, AxisError
    assert_raises(AxisError, io_axis_indices, cmap_b, 'q')
    # 0 leads to no match if fix0 set to false
    cmap_z = AffineTransform('ijk', 'xyz', np.diag([2, 3, 0, 1]))
    assert_equal(io_axis_indices(cmap_z, 'y'), (1, 1))
    assert_equal(io_axis_indices(cmap_z, 'z'), (2, 2))
    assert_equal(io_axis_indices(cmap_z, 'z', fix0=False), (None, 2))
    # For either input or output
    assert_equal(io_axis_indices(cmap_z, 'k'), (2, 2))
    assert_equal(io_axis_indices(cmap_z, 'k', fix0=False), (2, None))
    # axis name and number access without fix0
    cmap = AffineTransform('ijkt', 'xyzt', np.diag([1, 1, 1, 0, 1]))
    assert_raises(AxisError, io_axis_indices, cmap, 't', fix0=False)
    in_ax, out_ax = io_axis_indices(cmap, -1, fix0=False)
    assert_equal((in_ax, out_ax), (3, None))
    # Non-square is OK
    cmap = AffineTransform('ij', 'xyz', [[0, 1, 0],
                                         [0, 0, 0],
                                         [1, 0, 0],
                                         [0, 0, 1]])
    assert_equal(io_axis_indices(cmap, 'j'), (1, 0))
    assert_equal(io_axis_indices(cmap, 'y'), (None, 1))
    assert_equal(io_axis_indices(cmap, 'z'), (0, 2))
    cmap = AffineTransform('ijk', 'xy', [[0, 1, 0, 0],
                                         [0, 0, 1, 0],
                                         [0, 0, 0, 1]])
    assert_equal(io_axis_indices(cmap, 'i'), (0, None))
    assert_equal(io_axis_indices(cmap, 'j'), (1, 0))
    assert_equal(io_axis_indices(cmap, 'y'), (2, 1))


def test_make_cmap():
    # Routine to put the guessing back into making coordinate maps
    d_names = list('ijklm')
    r_names = list('xyztu')
    domain_maker = CoordSysMaker(d_names, 'voxels')
    range_maker = CoordSysMaker(r_names, 'world')
    cmm = CoordMapMaker(domain_maker, range_maker)
    # Making with generic functions and with affines
    xform = lambda x : x+1
    inv_xform = lambda x : x-1
    diag_vals = range(2,8)
    for i in range(1, 6):
        dcs = CS(d_names[:i], 'voxels')
        rcs = CS(r_names[:i], 'world')
        # Generic
        assert_equal(cmm.make_cmap(i, xform, inv_xform),
                     CoordinateMap(dcs, rcs, xform, inv_xform))
        assert_equal(cmm.make_cmap(i, xform), CoordinateMap(dcs, rcs, xform))
        # Affines
        aff = np.diag(diag_vals[:i] + [1])
        assert_equal(cmm.make_affine(aff), AffineTransform(dcs, rcs, aff))
        # Test that the call method selects what it got correctly
        assert_equal(cmm(i, xform, inv_xform),
                     CoordinateMap(dcs, rcs, xform, inv_xform))
        assert_equal(cmm(i, xform), CoordinateMap(dcs, rcs, xform))
        assert_equal(cmm(aff), AffineTransform(dcs, rcs, aff))
    # For affines, we can append dimensions by adding on the diagonal
    aff = np.diag([2,3,4,1])
    dcs = CS(d_names[:4], 'voxels')
    rcs = CS(r_names[:4], 'world')
    assert_equal(cmm.make_affine(aff, 5),
                 AffineTransform(CS(d_names[:4], 'voxels'),
                                 CS(r_names[:4], 'world'),
                                 np.diag([2,3,4,5,1])))
    assert_equal(cmm.make_affine(aff, [5,6]),
                 AffineTransform(CS(d_names[:5], 'voxels'),
                                 CS(r_names[:5], 'world'),
                                 np.diag([2,3,4,5,6,1])))
    # we can add offsets too
    exp_aff = np.diag([2,3,4,5,6,1])
    exp_aff[3:5,-1] = [7,8]
    assert_equal(cmm.make_affine(aff, [5,6],[7,8]),
                 AffineTransform(CS(d_names[:5], 'voxels'),
                                 CS(r_names[:5], 'world'),
                                 exp_aff))
    # The zooms (diagonal elements) and offsets must match in length
    assert_raises(CoordMapMakerError, cmm.make_affine, aff, [5,6], 7)
    # Check non-square affines
    aff = np.array([[2,0,0],
                    [0,3,0],
                    [0,0,1],
                    [0,0,1]])
    dcs = CS(d_names[:2], 'voxels')
    rcs = CS(r_names[:3], 'world')
    assert_equal(cmm.make_affine(aff), AffineTransform(dcs, rcs, aff))
    dcs = CS(d_names[:3], 'voxels')
    rcs = CS(r_names[:4], 'world')
    exp_aff = np.array([[2,0,0,0],
                        [0,3,0,0],
                        [0,0,0,1],
                        [0,0,4,0],
                        [0,0,0,1]])
    assert_equal(cmm.make_affine(aff, 4), AffineTransform(dcs, rcs, exp_aff))


def test_dtype_cmap_inverses():
    # Check that we can make functional inverses of AffineTransforms, and
    # CoordinateMap versions of AffineTransforms
    dtypes = (np.sctypes['int'] + np.sctypes['uint'] + np.sctypes['float']
              + np.sctypes['complex'] + [np.object])
    arr_p1 = np.eye(4)[:, [0, 2, 1, 3]]
    in_list = [0, 1, 2]
    out_list = [0, 2, 1]
    for dt in dtypes:
        in_cs = CoordinateSystem('ijk', coord_dtype=dt)
        out_cs = CoordinateSystem('xyz', coord_dtype=dt)
        cmap = AffineTransform(in_cs, out_cs, arr_p1.astype(dt))
        coord = np.array(in_list, dtype=dt)
        out_coord = np.array(out_list, dtype=dt)
        # Expected output type of inverse, not preserving
        if dt in np.sctypes['int'] + np.sctypes['uint']:
            exp_i_dt = np.float64
        else:
            exp_i_dt = dt
        # Default inverse cmap may alter coordinate types
        r_cmap = cmap.inverse()
        res = r_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, exp_i_dt)
        # Default behavior is preserve_type=False
        r_cmap = cmap.inverse(preserve_dtype=False)
        res = r_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, exp_i_dt)
        # Preserve_dtype=True - preserves dtype
        r_cmap = cmap.inverse(preserve_dtype=True)
        res = r_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, dt)
        # Preserve_dtype=True is default for conversion to CoordinateMap
        cm_cmap = _as_coordinate_map(cmap)
        assert_array_equal(cm_cmap(coord), out_list)
        rcm_cmap = cm_cmap.inverse()
        assert_array_equal(rcm_cmap(coord), out_list)
        res = rcm_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, dt)
    # For integer types, where there is no integer inverse, return floatey
    # inverse by default, and None for inverse when preserve_dtype=True
    arr_p2 = arr_p1 * 2
    arr_p2[-1, -1] = 1
    out_list = [0, 4, 2]
    for dt in np.sctypes['int'] + np.sctypes['uint']:
        in_cs = CoordinateSystem('ijk', coord_dtype=dt)
        out_cs = CoordinateSystem('xyz', coord_dtype=dt)
        cmap = AffineTransform(in_cs, out_cs, arr_p2.astype(dt))
        coord = np.array(in_list, dtype=dt)
        out_coord = np.array(out_list, dtype=dt)
        # Default
        r_cmap = cmap.inverse()
        res = r_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, np.float64)
        # Default is preserve_type=False
        r_cmap = cmap.inverse(preserve_dtype=False)
        res = r_cmap(out_coord)
        assert_array_equal(res, coord)
        assert_equal(res.dtype, np.float64)
        # preserve_dtype=True means there is no valid inverse for non integer
        # affine inverses, as here
        assert_equal(cmap.inverse(preserve_dtype=True), None)


def test_subtype_equalities():
    # Check cmap compare equal if subtypes, on either side
    in_cs = CoordinateSystem('ijk')
    out_cs = CoordinateSystem('xyz')
    f = lambda x : x + 1
    cmap = CoordinateMap(in_cs, out_cs, f)
    class CM2(CoordinateMap): pass
    cmap2 = CM2(in_cs, out_cs, f)
    assert_equal(cmap, cmap2)
    assert_equal(cmap2, cmap)
    cmap = AffineTransform(in_cs, out_cs, np.eye(4))
    class AT2(AffineTransform): pass
    cmap2 = AT2(in_cs, out_cs, np.eye(4))
    assert_equal(cmap, cmap2)
    assert_equal(cmap2, cmap)


def test_cmap_coord_types():
    # Check that we can use full range of coordinate system types.  The inverse
    # of an AffineTransform should generate coordinates in the input coordinate
    # system dtype
    dtypes = (np.sctypes['int'] + np.sctypes['uint'] + np.sctypes['float']
              + np.sctypes['complex'] + [np.object])
    arr_p1 = np.eye(4)
    arr_p1[:3, 3] = 1
    for dt in dtypes:
        in_cs = CoordinateSystem('ijk', coord_dtype=dt)
        out_cs = CoordinateSystem('xyz', coord_dtype=dt)
        # CoordinateMap
        cmap = CoordinateMap(in_cs, out_cs, lambda x : x + 1)
        assert_equal(cmap, copy(cmap))
        res = cmap(np.array([0, 1, 2], dtype=dt))
        assert_array_equal(res, [1, 2, 3])
        assert_equal(res.dtype, in_cs.coord_dtype)
        # Check reordering works
        rcmap = cmap.reordered_domain('ikj').reordered_range('yxz')
        res = rcmap(np.array([0, 1, 2], dtype=dt))
        assert_array_equal(res, [3, 1, 2])
        assert_equal(res.dtype, in_cs.coord_dtype)
        # AffineTransform
        cmap = AffineTransform(in_cs, out_cs, arr_p1.astype(dt))
        res = cmap(np.array([0, 1, 2], dtype=dt))
        assert_array_equal(res, [1, 2, 3])
        assert_equal(res.dtype, in_cs.coord_dtype)
        assert_equal(cmap, copy(cmap))
        # Check reordering works
        rcmap = cmap.reordered_domain('ikj').reordered_range('yxz')
        res = rcmap(np.array([0, 1, 2], dtype=dt))
        assert_array_equal(res, [3, 1, 2])
        assert_equal(res.dtype, in_cs.coord_dtype)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
__docformat__ = 'restructuredtext'

import numpy as np

class VarianceFunction(object):
    """
    Variance function that relates the variance of a random variable
    to its mean. Defaults to 1.
    """

    def __call__(self, mu):
        """
        Default variance function

        INPUTS:
           mu  -- mean parameters

        OUTPUTS: v
           v   -- ones(mu.shape)
        """

        return np.ones(mu.shape, np.float64)

constant = VarianceFunction()

class Power(object):
    """
    Power variance function:

    V(mu) = fabs(mu)**power

    INPUTS:
       power -- exponent used in power variance function

    """

    def __init__(self, power=1.):
        self.power = power

    def __call__(self, mu):

        """
        Power variance function

        INPUTS:
           mu  -- mean parameters

        OUTPUTS: v
           v   -- fabs(mu)**self.power
        """
        return np.power(np.fabs(mu), self.power)

class Binomial(object):
    """
    Binomial variance function

    p = mu / n; V(mu) = p * (1 - p) * n

    INPUTS:
       n -- number of trials in Binomial
    """

    tol = 1.0e-10

    def __init__(self, n=1):
        self.n = n

    def clean(self, p):
        return np.clip(p, Binomial.tol, 1 - Binomial.tol)

    def __call__(self, mu):
        """
        Binomial variance function

        INPUTS:
           mu  -- mean parameters

        OUTPUTS: v
           v   -- mu / self.n * (1 - mu / self.n) * self.n
        """
        p = self.clean(mu / self.n)
        return p * (1 - p) * self.n

mu = Power()
mu_squared = Power(power=2)
mu_cubed = Power(power=3)
binary = Binomial()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This module implements the Field class, which simply a WeightedGraph
(see the graph.py) module, plus an arrray that yields (possibly
multi-dimnesional) features associated with graph vertices. This
allows some kinds of computations (all thoses relating to mathematical
morphology, diffusion etc.)

Certain functions are provided to Instantiate Fields easily, given a
WeightedGraph and feature data.

Author:Bertrand Thirion, 2006--2011
"""
import numpy as np

from .graph import WeightedGraph

NEGINF = -np.inf


def field_from_coo_matrix_and_data(x, data):
    """ Instantiates a weighted graph from a (sparse) coo_matrix

    Parameters
    ----------
    x: (V, V) scipy.sparse.coo_matrix instance,
       the input matrix
    data: array of shape (V, dim),
          the field data

    Returns
    -------
    ifield: resulting Field instance
    """
    if x.shape[0] != x.shape[1]:
        raise ValueError("the input coo_matrix is not square")
    if data.shape[0] != x.shape[0]:
        raise ValueError("data and x do not have consistent shapes")
    i, j = x.nonzero()
    edges = np.vstack((i, j)).T
    weights = x.data
    ifield = Field(x.shape[0], edges, weights, data)
    return ifield


def field_from_graph_and_data(g, data):
    """ Instantiate a Fieldfrom a WeightedGraph plus some feature data
    Parameters
    ----------
    x: (V, V) scipy.sparse.coo_matrix instance,
       the input matrix
    data: array of shape (V, dim),
          the field data

    Returns
    -------
    ifield: resulting field instance
    """
    if data.shape[0] != g.V:
        raise ValueError("data and g do not have consistent shapes")
    ifield = Field(g.V, g.edges, g.weights, data)
    return ifield


class Field(WeightedGraph):
    """
    This is the basic field structure,
         which contains the weighted graph structure
         plus an array of data (the 'field')
    field is an array of size(n, p)
          where n is the number of vertices of the graph
          and p is the field dimension
    """

    def __init__(self, V, edges=None, weights=None, field=None):
        """
        Parameters
        ----------
        V (int > 0) the number of vertices of the graph
        edges=None: the edge array of the graph
        weights=None: the asociated weights array
        field=None: the field data itself
        """
        V = int(V)
        if V < 1:
            raise ValueError('cannot create graph with no vertex')
        self.V = int(V)
        self.E = 0
        self.edges = []
        self.weights = []
        if (edges is not None) or (weights is not None):
            if len(edges) == 0:
                E = 0
            elif edges.shape[0] == np.size(weights):
                E = edges.shape[0]
            else:
                raise ValueError('Incompatible size of the edges \
                                  and weights matrices')
            self.V = V
            self.E = E
            self.edges = edges
            self.weights = weights
        self.field = []
        if field == None:
            pass
        else:
            if np.size(field) == self.V:
                field = np.reshape(field, (self.V, 1))
            if field.shape[0] != self.V:
                raise ValueError('field does not have a correct size')
            else:
                self.field = field

    def get_field(self):
        return self.field

    def set_field(self, field):
        if np.size(field) == self.V:
            field = np.reshape(field, (self.V, 1))
        if field.shape[0] != self.V:
            raise ValueError('field does not have a correct size')
        else:
            self.field = field

    def closing(self, nbiter=1):
        """Morphological closing of the field data. 
        self.field is changed inplace

        Parameters
        ----------
        nbiter=1 : the number of iterations required
        """
        nbiter = int(nbiter)
        self.dilation(nbiter)
        self.erosion(nbiter)

    def opening(self, nbiter=1):
        """Morphological opening of the field data.
        self.field is changed inplace

        Parameters
        ----------
        nbiter: int, optional, the number of iterations required
        """
        nbiter = int(nbiter)
        self.erosion(nbiter)
        self.dilation(nbiter)

    def dilation(self, nbiter=1, fast=True):
        """
        Morphological dimlation of the field data. self.field is changed

        Parameters
        ----------
        nbiter: int, optional, the number of iterations required

        fixme
        -----
        cython
        """
        nbiter = int(nbiter)
        if fast:
            from ._graph import dilation
            if self.E > 0:
                if (self.field.size == self.V):
                    self.field = self.field.reshape((self.V, 1))
                idx, neighb, _ = self.compact_neighb()
                for i in range(nbiter):
                    dilation(self.field, idx, neighb)
        else:
            from scipy.sparse import dia_matrix
            adj = self.to_coo_matrix() + dia_matrix(
                (np.ones(self.V), 0), (self.V, self.V))
            rows = adj.tolil().rows
            for i in range(nbiter):
                self.field = np.array([self.field[row].max(0) for row in rows])

    def highest_neighbor(self, refdim=0):
        """ Computes the neighbor with highest field value along refdim

        Parameters
        ----------
        refdim: int optiontal, the dimension to consider

        Returns
        -------
        hneighb: array of shape(self.V), index of the neighbor with highest
                 value
        """
        from scipy.sparse import dia_matrix
        refdim = int(refdim)
        # add self-edges to avoid singularities, when taking the maximum
        adj = self.to_coo_matrix() + dia_matrix(
            (np.ones(self.V), 0), (self.V, self.V))
        rows = adj.tolil().rows
        hneighb = np.array([row[self.field[row].argmax()] for row in rows])
        return hneighb

    def erosion(self, nbiter=1):
        """Morphological openeing of the field

        Parameters
        ----------
        nbiter: int, optional, the number of iterations required
        """
        nbiter = int(nbiter)
        lil = self.to_coo_matrix().tolil().rows.tolist()
        for i in range(nbiter):
            nf = np.zeros_like(self.field)
            for k, neighbors in enumerate(lil):
                nf[k] = self.field[neighbors].min(0)
            self.field = nf

    def get_local_maxima(self, refdim=0, th=NEGINF):
        """
        Look for the local maxima of one dimension (refdim) of self.field

        Parameters
        ----------
        refdim (int) the field dimension over which the maxima are looked after
        th = float, optional
            threshold so that only values above th are considered

        Returns
        -------
        idx: array of shape (nmax)
             indices of the vertices that are local maxima
        depth: array of shape (nmax)
               topological depth of the local maxima :
               depth[idx[i]] = q means that idx[i] is a q-order maximum
        """
        depth_all = self.local_maxima(refdim, th)
        idx = np.ravel(np.where(depth_all))
        depth = depth_all[idx]
        return idx, depth

    def local_maxima(self, refdim=0, th=NEGINF):
        """Returns all the local maxima of a field

        Parameters
        ----------
        refdim (int) field dimension over which the maxima are looked after
        th: float, optional
            threshold so that only values above th are considered

        Returns
        -------
        depth: array of shape (nmax)
               a labelling of the vertices such that
               depth[v] = 0 if v is not a local maximum
               depth[v] = 1 if v is a first order maximum
               ...
               depth[v] = q if v is a q-order maximum
        """
        refdim = int(refdim)
        if np.size(self.field) == 0:
            raise ValueError('No field has been defined so far')
        if self.field.shape[1] - 1 < refdim:
            raise ValueError(refdim > self.shape[1])
        depth = np.zeros(self.V, np.int)

        # create a subfield(thresholding)
        sf = self.subfield(self.field.T[refdim] >= th)
        initial_field = sf.field.T[refdim]
        sf.field = initial_field.copy()

        # compute the depth in the subgraph
        ldepth = sf.V * np.ones(sf.V, np.int)
        for k in range(sf.V):
            dilated_field_old = sf.field.ravel().copy()
            sf.dilation(1)
            non_max = sf.field.ravel() > dilated_field_old
            ldepth[non_max] = np.minimum(k, ldepth[non_max])
            if (non_max == False).all():
                ldepth[sf.field.ravel() == initial_field] = np.maximum(k, 1)
                break

        # write all the depth values
        depth[self.field[:, refdim] >= th] = ldepth
        return depth

    def diffusion(self, nbiter=1):
        """diffusion of the field data in the weighted graph structure
        self.field is changed inplace

        Parameters
        ----------
        nbiter: int, optional the number of iterations required

        Notes
        -----
        The process is run for all the dimensions of the field
        """
        nbiter = int(nbiter)
        adj = self.to_coo_matrix()
        for i in range(nbiter):
            self.field = adj * self.field

    def custom_watershed(self, refdim=0, th=NEGINF):
        """ customized watershed analysis of the field.
        Note that bassins are found around each maximum
        (and not minimum as conventionally)

        Parameters
        ----------
        refdim: int, optional
        th: float optional, threshold of the field

        Returns
        -------
        idx: array of shape (nbassins)
             indices of the vertices that are local maxima
        label : array of shape (self.V)
              labelling of the vertices according to their bassin
        """
        import numpy.ma as ma
        from graph import Graph

        if (np.size(self.field) == 0):
            raise ValueError('No field has been defined so far')
        if self.field.shape[1] - 1 < refdim:
            raise ValueError('refdim>field.shape[1]')

        label = - np.ones(self.V, np.int)

        # create a subfield(thresholding)
        sf = self.subfield(self.field[:, refdim] >= th)

        # compute the basins
        hneighb = sf.highest_neighbor()
        edges = np.vstack((hneighb, np.arange(sf.V))).T
        edges = np.vstack((edges, np.vstack((np.arange(sf.V), hneighb)).T))
        aux = Graph(sf.V, edges.shape[0], edges)
        llabel = aux.cc()
        n_bassins = len(np.unique(llabel))

        # write all the depth values
        label[self.field[:, refdim] >= th] = llabel
        idx = np.array([ma.array(
                    self.field[:, refdim], mask=(label != c)).argmax()
                        for c in range(n_bassins)])
        return idx, label

    def threshold_bifurcations(self, refdim=0, th=NEGINF):
        """Analysis of the level sets of the field:
        Bifurcations are defined as changes in the topology in the level sets
        when the level (threshold) is varied
        This can been thought of as a kind of Morse analysis

        Parameters
        ----------
        th: float, optional,
            threshold so that only values above th are considered

        Returns
        -------
        idx: array of shape (nlsets)
             indices of the vertices that are local maxima
        height: array of shape (nlsets)
                the depth of the local maxima
                depth[idx[i]] = q means that idx[i] is a q-order maximum
                Note that this is also the diameter of the basins
                associated with local maxima
        parents: array of shape (nlsets)
                 the label of the maximum which dominates each local maximum
                 i.e. it describes the hierarchy of the local maxima
        label: array of shape (self.V)
               a labelling of thevertices according to their bassin
        """
        import numpy.ma as ma
        if (np.size(self.field) == 0):
            raise ValueError('No field has been defined so far')
        if self.field.shape[1] - 1 < refdim:
            raise ValueError('refdim>field.shape[1]')

        label = - np.ones(self.V, np.int)

        # create a subfield(thresholding)
        sf = self.subfield(self.field[:, refdim] >= th)
        initial_field = sf.field[:, refdim].copy()
        sf.field = initial_field.copy()

        # explore the subfield
        order = np.argsort(- initial_field)
        rows = sf.to_coo_matrix().tolil().rows
        llabel = - np.ones(sf.V, np.int)
        parent, root =  np.arange(2 * self.V), np.arange(2 * self.V)
        # q will denote the region index
        q = 0
        for i in order:
            if (llabel[rows[i]] > - 1).any():
                nlabel = np.unique(llabel[rows[i]])
                if nlabel[0] == -1:
                    nlabel = nlabel[1:]
                nlabel = np.unique(root[nlabel])
                if len(nlabel) == 1:
                    # we are at a regular point
                    llabel[i] = nlabel[0]
                else:
                    # we are at a saddle point
                    llabel[i] = q
                    parent[nlabel] = q
                    root[nlabel] = q
                    for j in nlabel:
                        root[root == j] = q
                    q += 1
            else:
                # this is a new component
                llabel[i] = q
                q += 1
        parent = parent[:q]

        # write all the depth values
        label[self.field[:, refdim] >= th] = llabel
        idx = np.array([ma.array(
                    self.field[:, refdim], mask=(label != c)).argmax()
                         for c in range(q)])
        return idx, parent, label

    def constrained_voronoi(self, seed):
        """Voronoi parcellation of the field starting from the input seed

        Parameters
        ----------
        seed: int array of shape(p), the input seeds

        Returns
        -------
        label: The resulting labelling of the data

        Fixme
        -----
        deal with graphs with several ccs
        """
        if np.size(self.field) == 0:
            raise ValueError('No field has been defined so far')
        seed = seed.astype(np.int)
        weights = np.sqrt(np.sum((self.field[self.edges.T[0]] -
                                  self.field[self.edges.T[1]]) ** 2, 1))
        g = WeightedGraph(self.V, self.edges, weights)
        label = g.voronoi_labelling(seed)
        return label

    def geodesic_kmeans(self, seeds=None, label=None, maxiter=100, eps=1.e-4,
                        verbose=0):
        """ Geodesic k-means algorithm
        i.e. obtention of clusters that are topologically
        connected and minimally variable concerning the information
        of self.field

        Parameters
        ----------
        seeds: array of shape(p), optional,
               initial indices of the seeds within the field
               if seeds==None the labels are used as initialization
        labels: array of shape(self.V) initial labels, optional,
                it is expected that labels take their values
                in a certain range (0..lmax)
                if Labels==None, this is not used
                if seeds==None and labels==None,  an ewxception is raised
        maxiter: int, optional,
                 maximal number of iterations
        eps: float, optional,
             increase of inertia at which convergence is declared

        Returns
        -------
        seeds: array of shape (p), the final seeds
        label : array of shape (self.V), the resulting field label
        J: float, inertia value
        """
        if np.size(self.field) == 0:
            raise ValueError('No field has been defined so far')

        if (seeds == None) and (label == None):
            raise ValueError('No initialization has been provided')
        k = np.size(seeds)
        inertia_old = NEGINF
        if seeds == None:
            k = label.max() + 1
            if np.size(np.unique(label)) != k:
                raise ValueError('missing values, cannot proceed')
            seeds = np.zeros(k).astype(np.int)
            for  j in range(k):
                lj = np.nonzero(label == j)[0]
                cent = np.mean(self.field[lj], 0)
                tj = np.argmin(np.sum((cent - self.field[lj]) ** 2, 1))
                seeds[j] = lj[tj]
        else:
            k = np.size(seeds)

        for i in range(maxiter):
            # voronoi labelling
            label = self.constrained_voronoi(seeds)
            # update the seeds
            inertia = 0
            pinteria = 0
            for  j in range(k):
                lj = np.nonzero(label == j)[0]
                pinteria += np.sum(
                    (self.field[seeds[j]] - self.field[lj]) ** 2)
                cent = np.mean(self.field[lj], 0)
                tj = np.argmin(np.sum((cent - self.field[lj]) ** 2, 1))
                seeds[j] = lj[tj]
                inertia += np.sum((cent - self.field[lj]) ** 2)
            if verbose:
                print i, inertia
            if np.absolute(inertia_old - inertia) < eps:
                break
            inertia_old = inertia
        return seeds, label, inertia

    def ward(self, nbcluster):
        """Ward's clustering of self

        Parameters
        ----------
        nbcluster: int,
                   the number of desired clusters

        Returns
        -------
        label: array of shape (self.V)
               the resulting field label
        J (float): the resulting inertia
        """
        from nipy.algorithms.clustering.hierarchical_clustering\
             import ward_segment
        label, J = ward_segment(self, self.field, qmax=nbcluster)

        # compute the resulting inertia
        inertia = 0
        for  j in range(nbcluster):
            lj = np.nonzero(label == j)[0]
            cent = np.mean(self.field[lj], 0)
            inertia += np.sum((cent - self.field[lj]) ** 2)
        return label, inertia

    def copy(self):
        """ copy function
        """
        return Field(self.V, self.edges.copy(),
                     self.weights.copy(), self.field.copy())

    def subfield(self, valid):
        """Returns a subfield of self, with only vertices such that valid > 0

        Parameters
        ----------
        valid: array of shape (self.V),
               nonzero for vertices to be retained

        Returns
        -------
        F: Field instance,
           the desired subfield of self

        Notes
        -----
        The vertices are renumbered as [1..p] where p = sum(valid>0) when
        sum(valid) == 0 then None is returned
        """
        G = self.subgraph(valid)
        if G == None:
            return None
        field = self.field[valid]
        if len(G.edges) == 0:
            edges = np.array([[], []]).T
        else:
            edges = G.edges
        return Field(G.V, edges, G.weights, field)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test scripts

If we appear to be running from the development directory, use the scripts in
the top-level folder ``scripts``.  Otherwise try and get the scripts from the
path
"""
from __future__ import with_statement

import sys
import os
from os.path import dirname, join as pjoin, isfile, isdir, abspath, realpath

from subprocess import Popen, PIPE

from nibabel.tmpdirs import InTemporaryDirectory

from nipy import load_image

from nose.tools import assert_true, assert_false, assert_equal

from ..testing import funcfile
from numpy.testing import decorators, assert_almost_equal

from nipy.testing.decorators import make_label_dec

from nibabel.optpkg import optional_package

matplotlib, HAVE_MPL, _ = optional_package('matplotlib')
needs_mpl = decorators.skipif(not HAVE_MPL, "Test needs matplotlib")
script_test = make_label_dec('script_test')

# Need shell to get path to correct executables
USE_SHELL = True

DEBUG_PRINT = os.environ.get('NIPY_DEBUG_PRINT', False)

def local_script_dir():
    # Check for presence of scripts in development directory.  ``realpath``
    # checks for the situation where the development directory has been linked
    # into the path.
    below_nipy_dir = realpath(pjoin(dirname(__file__), '..', '..'))
    devel_script_dir = pjoin(below_nipy_dir, 'scripts')
    if isfile(pjoin(below_nipy_dir, 'setup.py')) and isdir(devel_script_dir):
        return devel_script_dir
    return None

LOCAL_SCRIPT_DIR = local_script_dir()

def run_command(cmd):
    if not LOCAL_SCRIPT_DIR is None:
        # Windows can't run script files without extensions natively so we need
        # to run local scripts (no extensions) via the Python interpreter.  On
        # Unix, we might have the wrong incantation for the Python interpreter
        # in the hash bang first line in the source file.  So, either way, run
        # the script through the Python interpreter
        cmd = "%s %s" % (sys.executable, pjoin(LOCAL_SCRIPT_DIR, cmd))
    if DEBUG_PRINT:
        print("Running command '%s'" % cmd)
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=USE_SHELL)
    stdout, stderr = proc.communicate()
    if proc.poll() == None:
        proc.terminate()
    if proc.returncode != 0:
        raise RuntimeError('Command "%s" failed with stdout\n%s\nstderr\n%s\n'
                           % (cmd, stdout, stderr))
    return proc.returncode


@needs_mpl
@script_test
def test_nipy_diagnose():
    # Test nipy diagnose script
    fimg = load_image(funcfile)
    ncomps = 12
    with InTemporaryDirectory() as tmpdir:
        # Need to quote out path in case it has spaces
        cmd = 'nipy_diagnose "%s" --ncomponents=%d --out-path="%s"' % (
            funcfile, ncomps, tmpdir)
        run_command(cmd)
        for out_fname in ('components_functional.png',
                          'pcnt_var_functional.png',
                          'tsdiff_functional.png',
                          'vectors_components_functional.npz'):
            assert_true(isfile(out_fname))
        for out_img in ('max_functional.nii.gz',
                        'mean_functional.nii.gz',
                        'min_functional.nii.gz',
                        'std_functional.nii.gz'):
            img = load_image(out_img)
            assert_equal(img.shape, fimg.shape[:-1])
            del img
        pca_img = load_image('pca_functional.nii.gz')
        assert_equal(pca_img.shape, fimg.shape[:-1] + (ncomps,))
        del pca_img


@needs_mpl
@script_test
def test_nipy_tsdiffana():
    # Test nipy_tsdiffana script
    out_png = 'ts_out.png'
    with InTemporaryDirectory():
        # Quotes in case of space in arguments
        cmd = 'nipy_tsdiffana "%s" --out-file="%s"' % (funcfile, out_png)
        run_command(cmd)
        assert_true(isfile(out_png))


@script_test
def test_nipy_3_4d():
    # Test nipy_3dto4d and nipy_4dto3d
    fimg = load_image(funcfile)
    N = fimg.shape[-1]
    out_4d = 'func4d.nii'
    with InTemporaryDirectory() as tmpdir:
        # Quotes in case of space in arguments
        cmd = 'nipy_4dto3d "%s" --out-path="%s"' % (funcfile, tmpdir)
        run_command(cmd)
        imgs_3d = ['functional_%04d.nii' % i for i in range(N)]
        for iname in imgs_3d:
            assert_true(isfile(iname))
        cmd = 'nipy_3dto4d "%s" --out-4d="%s"' % ('" "'.join(imgs_3d), out_4d)
        run_command(cmd)
        fimg_back = load_image(out_4d)
        assert_almost_equal(fimg.get_data(), fimg_back.get_data())
        del fimg_back



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import numpy as np
from scipy.ndimage import affine_transform, map_coordinates

from ...core.image.image_spaces import (make_xyz_image,
                                        as_xyz_image,
                                        xyz_affine)
from .affine import inverse_affine, Affine
from ._registration import (_cspline_transform,
                            _cspline_sample3d,
                            _cspline_resample3d)


INTERP_ORDER = 3


def resample(moving, transform=None, reference=None,
             mov_voxel_coords=False, ref_voxel_coords=False,
             dtype=None, interp_order=INTERP_ORDER):
    """ Resample `movimg` into voxel space of `reference` using `transform`

    Apply a transformation to the image considered as 'moving' to
    bring it into the same grid as a given `reference` image.  The
    transformation usually maps world space in `reference` to world space in
    `movimg`, but can also be a voxel to voxel mapping (see parameters below).

    This function uses scipy.ndimage except for the case `interp_order==3`,
    where a fast cubic spline implementation is used.

    Parameters
    ----------
    moving: nipy-like image
      Image to be resampled.
    transform: transform object or None
      Represents a transform that goes from the `reference` image to
      the `moving` image. None means an identity transform. Otherwise,
      it should have either an `apply` method, or an `as_affine`
      method. By default, `transform` maps between the output (world)
      space of `reference` and the output (world) space of `moving`.
      If `mov_voxel_coords` is True, maps to the *voxel* space of
      `moving` and if `ref_vox_coords` is True, maps from the *voxel*
      space of `reference`.
    reference : None or nipy-like image or tuple, optional
      The reference image defines the image dimensions and xyz affine
      to which to resample. It can be input as a nipy-like image or as
      a tuple (shape, affine). If None, use `movimg` to define these.
    mov_voxel_coords : boolean, optional
      True if the transform maps to voxel coordinates, False if it
      maps to world coordinates.
    ref_voxel_coords : boolean, optional
      True if the transform maps from voxel coordinates, False if it
      maps from world coordinates.
    interp_order: int, optional
      Spline interpolation order, defaults to 3.

    Returns
    -------
    aligned_img : Image
        Image resliced to `reference` with reference-to-movimg transform
        `transform`
    """
    # Function assumes xyz_affine for inputs
    moving = as_xyz_image(moving)
    mov_aff = xyz_affine(moving)
    if reference == None:
        reference = moving
    if isinstance(reference, (tuple, list)):
        ref_shape, ref_aff = reference
    else:
        # Expecting image. Must be an image that can make an xyz_affine
        reference = as_xyz_image(reference)
        ref_shape = reference.shape
        ref_aff = xyz_affine(reference)
    if not len(ref_shape) == 3 or not ref_aff.shape == (4, 4):
        raise ValueError('Input image should be 3D')
    data = moving.get_data()
    if dtype == None:
        dtype = data.dtype

    # Assume identity transform by default
    if transform == None:
        transform = Affine()

    # Case: affine transform
    if hasattr(transform, 'as_affine'):
        Tv = transform.as_affine()
        if not ref_voxel_coords:
            Tv = np.dot(Tv, ref_aff)
        if not mov_voxel_coords:
            Tv = np.dot(inverse_affine(mov_aff), Tv)
        if interp_order == 3:
            output = _cspline_resample3d(data, ref_shape,
                                         Tv, dtype=dtype)
            output = output.astype(dtype)
        else:
            output = np.zeros(ref_shape, dtype=dtype)
            affine_transform(data, Tv[0:3, 0:3], offset=Tv[0:3, 3],
                             order=interp_order, cval=0,
                             output_shape=ref_shape, output=output)

    # Case: non-affine transform
    else:
        Tv = transform
        if not ref_voxel_coords:
            Tv = Tv.compose(Affine(ref_aff))
        if not mov_voxel_coords:
            Tv = Affine(inverse_affine(mov_aff)).compose(Tv)
        coords = np.indices(ref_shape).transpose((1, 2, 3, 0))
        coords = np.reshape(coords, (np.prod(ref_shape), 3))
        coords = Tv.apply(coords).T
        if interp_order == 3:
            cbspline = _cspline_transform(data)
            output = np.zeros(ref_shape, dtype='double')
            output = _cspline_sample3d(output, cbspline, *coords)
            output = output.astype(dtype)
        else:
            output = map_coordinates(data, coords, order=interp_order,
                                     cval=0, output=dtype)

    return make_xyz_image(output, ref_aff, 'scanner')



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
Simple demo that partitions a smooth field into 10 clusters.  In most cases,
Ward's clustering behaves best.

Requires matplotlib

Author: Bertrand Thirion, 2009
"""
print(__doc__)

import numpy as np
import numpy.random as nr

from scipy.ndimage import gaussian_filter

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nipy.algorithms.graph.field import Field

dx = 50
dy = 50
dz = 1
nbseeds = 10
data = gaussian_filter( np.random.randn(dx, dy), 2)
F = Field(dx * dy * dz)
xyz = np.reshape(np.indices((dx, dy, dz)), (3, dx * dy * dz)).T.astype(np.int)
F.from_3d_grid(xyz, 6)
F.set_field(data)

seeds = np.argsort(nr.rand(F.V))[:nbseeds]
seeds, label, J0 = F.geodesic_kmeans(seeds)
wlabel, J1 = F.ward(nbseeds)
seeds, label, J2 = F.geodesic_kmeans(seeds, label=wlabel.copy(), eps=1.e-7)

print('Inertia values for the 3 algorithms: ')
print('Geodesic k-means: ', J0, 'Wards: ', J1, 'Wards + gkm: ', J2)

plt.figure(figsize=(8, 4))
plt.subplot(1, 3, 1)
plt.imshow(np.reshape(data, (dx, dy)), interpolation='nearest')
plt.title('Input data')
plt.subplot(1, 3, 2)
plt.imshow(np.reshape(wlabel, (dx, dy)), interpolation='nearest')
plt.title('Ward clustering \n into 10 components')
plt.subplot(1, 3, 3)
plt.imshow(np.reshape(label, (dx, dy)), interpolation='nearest')
plt.title('geodesic kmeans clust. \n into 10 components')
plt.show()



''' Utilities for transforms3d '''

import math

import numpy as np


def normalized_vector(vec):
    ''' Return vector divided by Euclidean (L2) norm

    See :term:`unit vector` and :term:`Euclidean norm`

    Parameters
    ----------
    vec : array-like shape (3,)

    Returns
    -------
    nvec : array shape (3,)
       vector divided by L2 norm

    Examples
    --------
    >>> vec = [1, 2, 3]
    >>> l2n = np.sqrt(np.dot(vec, vec))
    >>> nvec = normalized_vector(vec)
    >>> np.allclose(np.array(vec) / l2n, nvec)
    True
    >>> vec = np.array([[1, 2, 3]])
    >>> vec.shape
    (1, 3)
    >>> normalized_vector(vec).shape
    (3,)
    '''
    vec = np.asarray(vec).squeeze()
    return vec / math.sqrt((vec**2).sum())


def vector_norm(vec):
    ''' Return vector Euclidaan (L2) norm

    See :term:`unit vector` and :term:`Euclidean norm`

    Parameters
    ----------
    vec : array-like shape (3,)

    Returns
    -------
    norm : scalar

    Examples
    --------
    >>> vec = [1, 2, 3]
    >>> l2n = np.sqrt(np.dot(vec, vec))
    >>> nvec = vector_norm(vec)
    >>> np.allclose(nvec, np.sqrt(np.dot(vec, vec)))
    True
    '''
    vec = np.asarray(vec)
    return math.sqrt((vec**2).sum())


def permutations(iterable, r=None):
    ''' Generate all permutations of `iterable`, of length `r`

    From Python docs, expressing 2.6 ``itertools.permutations``
    algorithm.

    If the elements are unique, then the resulting permutations will
    also be unique.

    Parameters
    ----------
    iterable : iterable
       returning elements that will be permuted
    r : None or int
       length of sequence to return, if None (default) use length of
       `iterable`

    Returns
    -------
    gen : generator
       generator that yields permutations
    
    Examples
    --------
    >>> tuple(permutations(range(3)))
    ((0, 1, 2), (0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1), (2, 1, 0))
    >>> tuple(permutations(range(3), 2))
    ((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))
    '''
    pool = tuple(iterable)
    n = len(pool)
    r = n if r is None else r
    if r > n:
        return
    indices = range(n)
    cycles = range(n, n-r, -1)
    yield tuple(pool[i] for i in indices[:r])
    while n:
        for i in reversed(range(r)):
            cycles[i] -= 1
            if cycles[i] == 0:
                indices[i:] = indices[i+1:] + indices[i:i+1]
                cycles[i] = n - i
            else:
                j = cycles[i]
                indices[i], indices[-j] = indices[-j], indices[i]
                yield tuple(pool[i] for i in indices[:r])
                break
        else:
            return


def inique(iterable):
    ''' Generate unique elements from `iterable`

    Parameters
    ----------
    iterable : iterable

    Returns
    -------
    gen : generator
       generator that yields unique elements from `iterable`
    
    Examples
    --------
    >>> tuple(inique([0, 1, 2, 0, 2, 3]))
    (0, 1, 2, 3)
    '''
    history = []
    for val in iterable:
        if val not in history:
            history.append(val)
            yield val


def permuted_signs(seq):
    ''' Generate permuted signs for sequence `seq`

    Parameters
    ----------
    seq : sequence

    Returns
    -------
    gen : generator
       generator returning `seq` with signs permuted

    Examples
    --------
    >>> tuple(permuted_signs([1, -2, 0]))
    ((1, -2, 0), (1, -2, 0), (1, 2, 0), (1, 2, 0), (-1, -2, 0), (-1, -2, 0), (-1, 2, 0), (-1, 2, 0))
    '''
    seq = tuple(seq)
    n = len(seq)
    for fs in inique(permutations([1]*n + [-1]*n, n)):
        yield tuple(e * f for e, f in zip(seq, fs))


def permuted_with_signs(seq):
    ''' Return all permutations of `seq` with all sign permutations

    Parameters
    ----------
    seq : sequence

    Returns
    -------
    gen : generator
       generator returning permutations and sign permutations

    Examples
    --------
    >>> tuple(permuted_with_signs((1,2)))
    ((1, 2), (1, -2), (-1, 2), (-1, -2), (2, 1), (2, -1), (-2, 1), (-2, -1))
    '''
    for pseq in permutations(seq):
        for sseq in permuted_signs(pseq):
            yield sseq



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
"""
Script that performs a first-level analysis of the FIAC dataset.

See ``examples/fiac/fiac_example.py`` for another approach to this analysis.

Needs the *example data* package.

Also needs matplotlib

Author: Alexis Roche, Bertrand Thirion, 2009--2012
"""

from os import mkdir, getcwd, path

import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nibabel import save

from nipy.modalities.fmri.glm import FMRILinearModel
from nipy.utils import example_data
from nipy.labs.viz import plot_map, cm

# -----------------------------------------------------------
# --------- Get the data -----------------------------------
#-----------------------------------------------------------

fmri_files = [example_data.get_filename('fiac', 'fiac0', run)
              for run in ['run1.nii.gz', 'run2.nii.gz']]
design_files = [example_data.get_filename('fiac', 'fiac0', run)
                for run in ['run1_design.npz', 'run2_design.npz']]
mask_file = example_data.get_filename('fiac', 'fiac0', 'mask.nii.gz')

# Load all the data
multi_session_model = FMRILinearModel(fmri_files, design_files, mask_file)

# GLM fitting
multi_session_model.fit(do_scaling=True, model='ar1')


def make_fiac_contrasts(p):
    """Specify some contrasts for the FIAC experiment

    Parameters
    ==========
    p: int, the number of columns of the design matrix (for all sessions)
    """
    con = {}
    # the design matrices of both runs comprise 13 columns
    # the first 5 columns of the design matrices correspond to the following
    # conditions: ["SSt-SSp", "SSt-DSp", "DSt-SSp", "DSt-DSp", "FirstSt"]

    def length_p_vector(con, p):
        return np.hstack((con, np.zeros(p - len(con))))

    con["SStSSp_minus_DStDSp"] = length_p_vector([1, 0, 0, - 1], p)
    con["DStDSp_minus_SStSSp"] = length_p_vector([- 1, 0, 0, 1], p)
    con["DSt_minus_SSt"] = length_p_vector([- 1, - 1, 1, 1], p)
    con["DSp_minus_SSp"] = length_p_vector([- 1, 1, - 1, 1], p)
    con["DSt_minus_SSt_for_DSp"] = length_p_vector([0, - 1, 0, 1], p)
    con["DSp_minus_SSp_for_DSt"] = length_p_vector([0, 0, - 1, 1], p)
    con["Deactivation"] = length_p_vector([- 1, - 1, - 1, - 1, 4], p)
    con["Effects_of_interest"] = np.eye(p)[:5]
    return con


# compute fixed effects of the two runs and compute related images
n_regressors = np.load(design_files[0])['X'].shape[1]
# note: implictly assume the same shape for all sessions !
contrasts = make_fiac_contrasts(n_regressors)

# write directory
write_dir = path.join(getcwd(), 'results')
if not path.exists(write_dir):
    mkdir(write_dir)

print('Computing contrasts...')
mean_map = multi_session_model.means[0]  # for display
for index, (contrast_id, contrast_val) in enumerate(contrasts.items()):
    print('  Contrast % 2i out of %i: %s' % (
        index + 1, len(contrasts), contrast_id))
    z_image_path = path.join(write_dir, '%s_z_map.nii' % contrast_id)
    z_map, = multi_session_model.contrast(
        [contrast_val] * 2, con_id=contrast_id, output_z=True)
    save(z_map, z_image_path)

    # make a snapshot of the contrast activation
    if contrast_id == 'Effects_of_interest':
        vmax = max(- z_map.get_data().min(), z_map.get_data().max())
        vmin = - vmax
        plot_map(z_map.get_data(), z_map.get_affine(),
                 anat=mean_map.get_data(), anat_affine=mean_map.get_affine(),
                 cmap=cm.cold_hot,
                 vmin=vmin,
                 vmax=vmax,
                 figure=10,
                 threshold=2.5,
                 black_bg=True)
        plt.savefig(path.join(write_dir, '%s_z_map.png' % contrast_id))

print("All the  results were witten in %s" % write_dir)
plt.show()



#!/usr/bin/env python

#
# Test fff linear algebra routines 
#

from numpy.testing import assert_equal
import numpy as np
from .. import vector_get, vector_set 

n = 15

def test_vector_get():
    x = np.random.rand(n)
    i = np.random.randint(n)
    xi = vector_get(x, i)
    assert_equal(xi, x[i])

def test_vector_get_int32():
    x = (100*np.random.rand(n)).astype('int32')
    i = np.random.randint(n)
    xi = vector_get(x, i)
    assert_equal(xi, x[i])

def test_vector_set():
    x = np.random.rand(n)
    i = np.random.randint(n)
    y = vector_set(x, i, 3)
    assert_equal(3, y[i])
    
def test_vector_set_int32():
    x = (100*np.random.rand(n)).astype('int32')
    i = np.random.randint(n)
    y = vector_set(x, i, 3)
    assert_equal(3, y[i])



if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])




"""
Compatibility module
"""

import warnings
warnings.warn(DeprecationWarning(
    "This module (nipy.labs.utils.mask) has been moved and "
    "is depreciated. Please update your code to import from "
    "'nipy.labs.mask'."))

# Absolute import, as 'import *' doesnot work with relative imports
from nipy.labs.mask import *



#!/usr/bin/env python
""" Refresh modification times on Cython generated C files

This is sometimes necessary on windows when the git checkout appears to
sometimes checkout the C files with modification times earlier than the pyx
files, triggering an attempt to rebuild the C files with Cython when running a
build.
"""
from __future__ import with_statement
import os
from os.path import splitext, join as pjoin, isfile
import sys
import optparse


# From http://stackoverflow.com/questions/1158076/implement-touch-using-python
if sys.version_info[0] >= 3:
    def touch(fname, times=None, ns=None, dir_fd=None):
        with os.open(fname, os.O_APPEND, dir_fd=dir_fd) as f:
            os.utime(f.fileno() if os.utime in os.supports_fd else fname,
                times=times, ns=ns, dir_fd=dir_fd)
else:
    def touch(fname, times=None):
        with file(fname, 'a'):
            os.utime(fname, times)


def main():
    parser = optparse.OptionParser(usage='%prog [<root_dir>]')
    (opts, args) = parser.parse_args()
    if len(args) > 1:
        parser.print_help()
        sys.exit(-1)
    elif len(args) == 1:
        root_dir = args[0]
    else:
        root_dir = os.getcwd()
    for dirpath, dirnames, filenames in os.walk(root_dir):
        for fn in filenames:
            if fn.endswith('.pyx'):
                froot, ext = splitext(fn)
                cfile = pjoin(dirpath, froot + '.c')
                if isfile(cfile):
                    touch(cfile)


if __name__ == '__main__':
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The volume data class

This class represents indexable data embedded in a 3D space 
"""

import copy as copy

import numpy as np

# Local imports
from .volume_field import VolumeField
from ..transforms.transform import CompositionError


################################################################################
# class `VolumeData`
################################################################################

class VolumeData(VolumeField):
    """ A class representing data embedded in a 3D space 

        This object has data stored in an array like, that knows how it is 
        mapped to a 3D "real-world space", and how it can change real-world 
        coordinate system.

        Attributes
        -----------

        world_space: string 
            World space the data is embedded in. For instance `mni152`.
        metadata: dictionnary
            Optional, user-defined, dictionnary used to carry around
            extra information about the data as it goes through
            transformations. The class consistency of this information is
            not maintained as the data is modified.
        _data: 
            Private pointer to the data.

        Notes
        ------

        The data is stored in an undefined way: prescalings might need to
        be applied to it before using it, or the data might be loaded on
        demand. The best practice to access the data is not to access the
        _data attribute, but to use the `get_data` method.
    """
    #---------------------------------------------------------------------------
    # Public attributes -- VolumeData interface
    #---------------------------------------------------------------------------

    # The interpolation logic used
    interpolation = 'continuous'

    #---------------------------------------------------------------------------
    # Private attributes -- VolumeData interface
    #---------------------------------------------------------------------------

    # The data (ndarray-like)
    _data = None

    #---------------------------------------------------------------------------
    # Public methods -- VolumeData interface
    #---------------------------------------------------------------------------


    def get_data(self):
        """ Return data as a numpy array.
        """
        return np.asanyarray(self._data)


    def like_from_data(self, data):
        """ Returns an volumetric data structure with the same
            relationship between data and world space, and same metadata, 
            but different data.

            Parameters
            -----------
            data: ndarray
        """
        raise NotImplementedError  


    def resampled_to_img(self, target_image, interpolation=None):
        """ Resample the data to be on the same voxel grid than the target 
            volume structure.

            Parameters
            ----------
            target_image : nipy image
                Nipy image onto the voxel grid of which the data will be
                resampled. This can be any kind of img understood by Nipy
                (datasets, pynifti objects, nibabel object) or a string
                giving the path to a nifti of analyse image.
            interpolation : None, 'continuous' or 'nearest', optional
                Interpolation type used when calculating values in
                different word spaces. If None, the image's interpolation
                logic is used.

            Returns
            -------
            resampled_image : nipy_image
                New nipy image with the data resampled.

            Notes
            -----
            Both the target image and the original image should be
            embedded in the same world space.
        """
        if not hasattr(target_image, 'world_space'):
            from ..converters import as_volume_img
            target_image = as_volume_img(target_image)
        if not target_image.world_space == self.world_space:
            raise CompositionError(
                "The two images are not embedded in the same world space")
        x, y, z = target_image.get_world_coords()
        new_data = self.values_in_world(x, y, z, 
                                        interpolation=interpolation)
        new_img = target_image.like_from_data(new_data) 
        new_img.metadata = copy.copy(self.metadata)
        return new_img


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------

    def _apply_transform(self, w2w_transform):
        """ Method applying the transform: inner part of
            transformed_with, used in subclassing.
        """
        new_v2w_transform = \
                        self.get_transform().composed_with(w2w_transform)
        new_img = copy.copy(self)
        new_img._transform = new_v2w_transform
        return new_img


    def _get_interpolation_order(self, interpolation):
        """ Inner method used to get the interpolation type for the
            image.
        """
        if interpolation is None:
            interpolation = self.interpolation
        if interpolation == 'continuous':
            interpolation_order = 3
        elif interpolation == 'nearest':
            interpolation_order = 0
        else:
            raise ValueError("interpolation must be either 'continuous' "
                             "or 'nearest'")
        return interpolation_order

    # TODO: We need to implement (or check if implemented) hashing,
    # weakref, pickling? 
        

    def __repr__(self):
        options = np.get_printoptions()
        np.set_printoptions(precision=5, threshold=64, edgeitems=2)
        representation = \
                '%s(\n  data=%s,\n  world_space=%s,\n  interpolation=%s)' % (
                self.__class__.__name__,
                '\n       '.join(repr(self._data).split('\n')),
                repr(self.world_space),
                repr(self.interpolation),
                )
        np.set_printoptions(**options)
        return representation


    def __copy__(self):
        return self.like_from_data(self.get_data().copy())


    def __deepcopy__(self, option):
        """ Copy the Volume and the arrays and metadata it contains.
        """
        out = self.__copy__()
        out.metadata = copy.deepcopy(self.metadata)
        return out


    def __eq__(self, other):
        return (    self.world_space       == other.world_space 
                and self.get_transform()   == other.get_transform()
                and np.all(self.get_data() == other.get_data())
                and self.interpolation     == other.interpolation
               )




"""
Infinite mixture model : A generalization of Bayesian mixture models
with an unspecified number of classes
"""
import numpy as np
from bgmm import BGMM, detsh
from scipy.special import gammaln


def co_labelling(z, kmax=None, kmin=None):
    """
    return a sparse co-labelling matrix given the label vector z

    Parameters
    ----------
    z: array of shape(n_samples),
       the input labels
    kmax: int, optional,
          considers only the labels in the range [0, kmax[

    Returns
    -------
    colabel: a sparse coo_matrix,
             yields the co labelling of the data
             i.e. c[i,j]= 1 if z[i]==z[j], 0 otherwise
    """
    from scipy.sparse import coo_matrix
    n = z.size
    colabel = coo_matrix((n, n))

    if kmax == None:
        kmax = z.max() + 1

    if kmin == None:
        kmin = z.min() - 1

    for  k in np.unique(z):
        if (k < kmax) & (k > kmin):
            i = np.array(np.nonzero(z == k))
            row = np.repeat(i, i.size)
            col = np.ravel(np.tile(i, i.size))
            data = np.ones((i.size) ** 2)
            colabel = colabel + coo_matrix((data, (row, col)), shape=(n, n))
    return colabel


class IMM(BGMM):
    """
    The class implements Infinite Gaussian Mixture model
    or Dirichlet Proces Mixture Model.
    This simply a generalization of Bayesian Gaussian Mixture Models
    with an unknown number of classes.
    """

    def __init__(self, alpha=.5, dim=1):
        """
        Parameters
        ----------
        alpha: float, optional,
               the parameter for cluster creation
        dim: int, optional,
             the dimension of the the data

        Note: use the function set_priors() to set adapted priors
        """
        self.dim = dim
        self.alpha = alpha
        self.k = 0
        self.prec_type = 'full'

        # initialize weights
        self.weights = [1]

    def set_priors(self, x):
        """ Set the priors in order of having them weakly uninformative
        this is from  Fraley and raftery;
        Journal of Classification 24:155-181 (2007)

        Parameters
        ----------
        x, array of shape (n_samples,self.dim)
           the data used in the estimation process
        """
        # a few parameters
        small = 0.01
        elshape = (1, self.dim, self.dim)
        mx = np.reshape(x.mean(0), (1, self.dim))
        dx = x - mx
        vx = np.maximum(1.e-15, np.dot(dx.T, dx) / x.shape[0])
        px = np.reshape(np.diag(1.0 / np.diag(vx)), elshape)

        # set the priors
        self._prior_means = mx
        self.prior_means = mx
        self.prior_weights = self.alpha
        self._prior_scale = px
        self.prior_scale = px
        self._prior_dof = self.dim + 2
        self.prior_dof = [self._prior_dof]
        self._prior_shrinkage = small
        self.prior_shrinkage = [self._prior_shrinkage]

        # cache some pre-computations
        self._dets_ = detsh(px[0])
        self._dets = [self._dets_]
        self._inv_prior_scale_ = np.reshape(np.linalg.inv(px[0]), elshape)
        self.prior_dens = None

    def set_constant_densities(self, prior_dens=None):
        """Set the null and prior densities as constant
        (assuming a compact domain)

        Parameters
        ----------
        prior_dens: float, optional
                    constant for the prior density
        """
        self.prior_dens = prior_dens

    def sample(self, x, niter=1, sampling_points=None, init=False,
               kfold=None, verbose=0):
        """sample the indicator and parameters

        Parameters
        ----------
        x: array of shape (n_samples, self.dim)
           the data used in the estimation process
        niter: int,
               the number of iterations to perform
        sampling_points: array of shape(nbpoints, self.dim), optional
                         points where the likelihood will be sampled
                         this defaults to x
        kfold: int or array, optional,
               parameter of cross-validation control
               by default, no cross-validation is used
               the procedure is faster but less accurate
        verbose=0: verbosity mode

        Returns
        -------
        likelihood: array of shape(nbpoints)
                    total likelihood of the model
        """
        self.check_x(x)
        if sampling_points == None:
            average_like = np.zeros(x.shape[0])
        else:
            average_like = np.zeros(sampling_points.shape[0])
            splike = self.likelihood_under_the_prior(sampling_points)

        plike = self.likelihood_under_the_prior(x)

        if init:
            self.k = 1
            z = np.zeros(x.shape[0])
            self.update(x, z)

        like = self.likelihood(x, plike)
        z = self.sample_indicator(like)

        for i in range(niter):
            if  kfold == None:
                like = self.simple_update(x, z, plike)
            else:
                like = self.cross_validated_update(x, z, plike, kfold)

            if sampling_points == None:
                average_like += like
            else:
                average_like += np.sum(
                    self.likelihood(sampling_points, splike), 1)

        average_like /= niter
        return average_like

    def simple_update(self, x, z, plike):
        """
         This is a step in the sampling procedure
        that uses internal corss_validation

        Parameters
        ----------
        x: array of shape(n_samples, dim),
           the input data
        z: array of shape(n_samples),
           the associated membership variables
        plike: array of shape(n_samples),
               the likelihood under the prior

        Returns
        -------
        like: array od shape(n_samples),
              the likelihood of the data
        """
        like = self.likelihood(x, plike)
        # standard + likelihood under the prior
        # like has shape (x.shape[0], self.k+1)

        z = self.sample_indicator(like)
        # almost standard, but many new components can be created

        self.reduce(z)
        self.update(x, z)
        return like.sum(1)

    def cross_validated_update(self, x, z, plike, kfold=10):
        """
        This is a step in the sampling procedure
        that uses internal corss_validation

        Parameters
        ----------
        x: array of shape(n_samples, dim),
           the input data
        z: array of shape(n_samples),
           the associated membership variables
        plike: array of shape(n_samples),
               the likelihood under the prior
        kfold: int, or array of shape(n_samples), optional,
               folds in the cross-validation loop

        Returns
        -------
        like: array od shape(n_samples),
              the (cross-validated) likelihood of the data
        """
        n_samples = x.shape[0]
        slike = np.zeros(n_samples)

        if np.isscalar(kfold):
            aux = np.argsort(np.random.rand(n_samples))
            idx = - np.ones(n_samples).astype(np.int)
            j = np.ceil(n_samples / kfold)
            kmax = kfold
            for k in range(kmax):
                idx[aux[k * j:min(n_samples, j * (k + 1))]] = k
        else:
            if np.array(kfold).size != n_samples:
                raise ValueError('kfold and x do not have the same size')
            uk = np.unique(kfold)
            np.random.shuffle(uk)
            idx = np.zeros(n_samples).astype(np.int)
            for i, k in enumerate(uk):
                idx += (i * (kfold == k))
            kmax = uk.max() + 1

        for k in range(kmax):
            test = np.zeros(n_samples).astype('bool')
            test[idx == k] = 1
            train = np.logical_not(test)

            # remove a fraction of the data
            # and re-estimate the clusters
            z[train] = self.reduce(z[train])
            self.update(x[train], z[train])

            # draw the membership for the left-out datas
            alike = self.likelihood(x[test], plike[test])
            slike[test] = alike.sum(1)
            # standard + likelihood under the prior
            # like has shape (x.shape[0], self.k+1)

            z[test] = self.sample_indicator(alike)
            # almost standard, but many new components can be created

        return slike

    def reduce(self, z):
        """Reduce the assignments by removing empty clusters and update self.k

        Parameters
        ----------
        z: array of shape(n),
           a vector of membership variables changed in place

        Returns
        -------
        z: the remapped values
        """
        uz = np.unique(z[z > - 1])
        for i, k in enumerate(uz):
            z[z == k] = i
        self.k = z.max() + 1
        return z

    def update(self, x, z):
        """ Update function (draw a sample of the IMM parameters)

        Parameters
        ----------
        x array of shape (n_samples,self.dim)
          the data used in the estimation process
        z array of shape (n_samples), type = np.int
          the corresponding classification
        """
        # re-dimension the priors in order to match self.k
        self.prior_means = np.repeat(self._prior_means, self.k, 0)
        self.prior_dof = self._prior_dof * np.ones(self.k)
        self.prior_shrinkage = self._prior_shrinkage * np.ones(self.k)
        self._dets = self._dets_ * np.ones(self.k)
        self._inv_prior_scale = np.repeat(self._inv_prior_scale_, self.k, 0)

        # initialize some variables
        self.means = np.zeros((self.k, self.dim))
        self.precisions = np.zeros((self.k, self.dim, self.dim))

        # proceed with the update
        BGMM.update(self, x, z)

    def update_weights(self, z):
        """
        Given the allocation vector z, resmaple the weights parameter

        Parameters
        ----------
        z array of shape (n_samples), type = np.int
          the allocation variable
        """
        pop = np.hstack((self.pop(z), 0))
        self.weights = pop + self.prior_weights
        self.weights /= self.weights.sum()

    def sample_indicator(self, like):
        """ Sample the indicator from the likelihood

        Parameters
        ----------
        like: array of shape (nbitem,self.k)
           component-wise likelihood

        Returns
        -------
        z: array of shape(nbitem): a draw of the membership variable

        Notes
        -----
        The behaviour is different from standard bgmm in that z can take
        arbitrary values
        """
        z = BGMM.sample_indicator(self, like)
        z[z == self.k] = self.k + np.arange(np.sum(z == self.k))
        return z

    def likelihood_under_the_prior(self, x):
        """ Computes the likelihood of x under the prior

        Parameters
        ----------
        x, array of shape (self.n_samples,self.dim)

        returns
        -------
        w, the likelihood of x under the prior model (unweighted)
        """
        if self.prior_dens is not None:
            return self.prior_dens * np.ones(x.shape[0])

        a = self._prior_dof
        tau = self._prior_shrinkage
        tau /= (1 + tau)
        m = self._prior_means
        b = self._prior_scale
        ib = np.linalg.inv(b[0])
        ldb = np.log(detsh(b[0]))

        scalar_w = np.log(tau / np.pi) * self.dim
        scalar_w += 2 * gammaln((a + 1) / 2)
        scalar_w -= 2 * gammaln((a - self.dim) / 2)
        scalar_w -= ldb * a
        w = scalar_w * np.ones(x.shape[0])

        for i in range(x.shape[0]):
            w[i] -= (a + 1) * np.log(detsh(ib + tau * (m - x[i:i + 1]) *
                                           (m - x[i:i + 1]).T))

        w /= 2
        return np.exp(w)

    def likelihood(self, x, plike=None):
        """
        return the likelihood of the model for the data x
        the values are weighted by the components weights

        Parameters
        ----------
        x: array of shape (n_samples, self.dim),
           the data used in the estimation process
        plike: array os shape (n_samples), optional,x
               the desnity of each point under the prior

        Returns
        -------
        like, array of shape(nbitem,self.k)
        component-wise likelihood
        """
        if plike == None:
            plike = self.likelihood_under_the_prior(x)

        plike = np.reshape(plike, (x.shape[0], 1))
        if self.k > 0:
            like = self.unweighted_likelihood(x)
            like = np.hstack((like, plike))
        else:
            like = plike
        like *= self.weights
        return like


class MixedIMM(IMM):
    """
    Particular IMM with an additional null class.
    The data is supplied together
    with a sample-related probability of being under the null.
    """

    def __init__(self, alpha=.5, dim=1):
        """
        Parameters
        ----------
        alpha: float, optional,
               the parameter for cluster creation
        dim: int, optional,
             the dimension of the the data

        Note: use the function set_priors() to set adapted priors
        """
        IMM.__init__(self, alpha, dim)

    def set_constant_densities(self, null_dens=None, prior_dens=None):
        """
        Set the null and prior densities as constant
        (over a  supposedly compact domain)

        Parameters
        ----------
        null_dens: float, optional
                   constant for the null density
        prior_dens: float, optional
                    constant for the prior density
        """
        self.null_dens = null_dens
        self.prior_dens = prior_dens

    def sample(self, x, null_class_proba, niter=1, sampling_points=None,
               init=False, kfold=None, co_clustering=False, verbose=0):
        """
        sample the indicator and parameters

        Parameters
        ----------
        x: array of shape (n_samples, self.dim),
           the data used in the estimation process
        null_class_proba: array of shape(n_samples),
                          the probability to be under the null
        niter: int,
               the number of iterations to perform
        sampling_points: array of shape(nbpoints, self.dim), optional
                         points where the likelihood will be sampled
                         this defaults to x
        kfold: int, optional,
               parameter of cross-validation control
               by default, no cross-validation is used
               the procedure is faster but less accurate
        co_clustering: bool, optional
                       if True,
                       return a model of data co-labelling across iterations
        verbose=0: verbosity mode

        Returns
        -------
        likelihood: array of shape(nbpoints)
                    total likelihood of the model
        pproba: array of shape(n_samples),
                the posterior of being in the null
                (the posterior of null_class_proba)
        coclust: only if co_clustering==True,
                 sparse_matrix of shape (n_samples, n_samples),
                 frequency of co-labelling of each sample pairs
                 across iterations
        """
        self.check_x(x)
        pproba = np.zeros(x.shape[0])

        if sampling_points == None:
            average_like = np.zeros(x.shape[0])
        else:
            average_like = np.zeros(sampling_points.shape[0])
            splike = self.likelihood_under_the_prior(sampling_points)

        plike = self.likelihood_under_the_prior(x)

        if init:
            self.k = 1
            z = np.zeros(x.shape[0])
            self.update(x, z)

        like = self.likelihood(x, plike)
        z = self.sample_indicator(like, null_class_proba)

        if co_clustering:
            from scipy.sparse import coo_matrix
            coclust = coo_matrix((x.shape[0], x.shape[0]))

        for i in range(niter):
            if  kfold == None:
                like = self.simple_update(x, z, plike, null_class_proba)
            else:
                like, z = self.cross_validated_update(x, z, plike,
                                                      null_class_proba, kfold)

            llike = self.likelihood(x, plike)
            z = self.sample_indicator(llike, null_class_proba)
            pproba += (z == - 1)

            if co_clustering:
                coclust = coclust + co_labelling(z, self.k, -1)

            if sampling_points == None:
                average_like += like
            else:
                average_like += np.sum(
                    self.likelihood(sampling_points, splike), 1)

        average_like /= niter
        pproba /= niter
        if co_clustering:
            coclust /= niter
            return average_like, pproba, coclust
        return average_like, pproba

    def simple_update(self, x, z, plike, null_class_proba):
        """ One step in the sampling procedure (one data sweep)

        Parameters
        ----------
        x: array of shape(n_samples, dim),
           the input data
        z: array of shape(n_samples),
           the associated membership variables
        plike: array of shape(n_samples),
               the likelihood under the prior
        null_class_proba: array of shape(n_samples),
                          prior probability to be under the null

        Returns
        -------
        like: array od shape(n_samples),
              the likelihood of the data under the H1 hypothesis
        """
        like = self.likelihood(x, plike)
        # standard + likelihood under the prior
        # like has shape (x.shape[0], self.k+1)

        z = self.sample_indicator(like, null_class_proba)
        # almost standard, but many new components can be created

        self.reduce(z)
        self.update(x, z)
        return like.sum(1)

    def cross_validated_update(self, x, z, plike, null_class_proba, kfold=10):
        """
        This is a step in the sampling procedure
        that uses internal corss_validation

        Parameters
        ----------
        x: array of shape(n_samples, dim),
           the input data
        z: array of shape(n_samples),
           the associated membership variables
        plike: array of shape(n_samples),
               the likelihood under the prior
        kfold: int, optional, or array
               number of folds in cross-validation loop
               or set of indexes for the cross-validation procedure
        null_class_proba: array of shape(n_samples),
                          prior probability to be under the null

        Returns
        -------
        like: array od shape(n_samples),
              the (cross-validated) likelihood of the data
        z: array of shape(n_samples),
              the associated membership variables

        Notes
        -----
        When kfold is an array, there is an internal reshuffling to randomize
        the order of updates
        """
        n_samples = x.shape[0]
        slike = np.zeros(n_samples)

        if np.isscalar(kfold):
            aux = np.argsort(np.random.rand(n_samples))
            idx = - np.ones(n_samples).astype(np.int)
            j = np.ceil(n_samples / kfold)
            kmax = kfold
            for k in range(kmax):
                idx[aux[k * j:min(n_samples, j * (k + 1))]] = k
        else:
            if np.array(kfold).size != n_samples:
                raise ValueError('kfold and x do not have the same size')
            uk = np.unique(kfold)
            np.random.shuffle(uk)
            idx = np.zeros(n_samples).astype(np.int)
            for i, k in enumerate(uk):
                idx += (i * (kfold == k))
            kmax = uk.max() + 1

        for k in range(kmax):
            # split at iteration k
            test = np.zeros(n_samples).astype('bool')
            test[idx == k] = 1
            train = np.logical_not(test)

            # remove a fraction of the data
            # and re-estimate the clusters
            z[train] = self.reduce(z[train])
            self.update(x[train], z[train])

            # draw the membership for the left-out data
            alike = self.likelihood(x[test], plike[test])
            slike[test] = alike.sum(1)
            # standard + likelihood under the prior
            # like has shape (x.shape[0], self.k+1)

            z[test] = self.sample_indicator(alike, null_class_proba[test])
            # almost standard, but many new components can be created

        return slike, z

    def sample_indicator(self, like, null_class_proba):
        """
        sample the indicator from the likelihood

        Parameters
        ----------
        like: array of shape (nbitem,self.k)
           component-wise likelihood
        null_class_proba: array of shape(n_samples),
                          prior probability to be under the null

        Returns
        -------
        z: array of shape(nbitem): a draw of the membership variable

        Notes
        -----
        Here z=-1 encodes for the null class
        """
        n = like.shape[0]
        conditional_like_1 = ((1 - null_class_proba) * like.T).T
        conditional_like_0 = np.reshape(null_class_proba *
                                        self.null_dens, (n, 1))
        conditional_like = np.hstack((conditional_like_0, conditional_like_1))
        z = BGMM.sample_indicator(self, conditional_like) - 1
        z[z == self.k] = self.k + np.arange(np.sum(z == self.k))
        return z


def main():
    """ Illustrative example of the behaviour of imm
    """
    n = 100
    dim = 2
    alpha = .5
    aff = np.random.randn(dim, dim)
    x = np.dot(np.random.randn(n, dim), aff)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100, kfold=10)
    print 'number of components: ', igmm.k

    #
    print 'number of components: ', igmm.k

    if dim < 3:
        from gmm import plot2D
        plot2D(x, igmm, verbose=1)
    return igmm


if __name__ == '__main__':
    main()



""" Nipy nosetester

Sets doctests to run by default

Use our own doctest plugin (based on that of numpy)
"""
from ..fixes.numpy.testing.nosetester import NoseTester, import_nose

def fpw_opt_str():
    """ Return first-package-wins option string for this version of nose

    Versions of nose prior to 1.1.0 needed ``=True`` for ``first-package-wins``,
    versions after won't accept it.

    changeset:   816:c344a4552d76
    http://code.google.com/p/python-nose/issues/detail?id=293

    Returns
    -------
    fpw_str : str
        Either '--first-package-wins' or '--first-package-wins=True' depending
        on the nose version we are running.
    """
    # protect nose import to provide comprehensible error if missing
    nose = import_nose()
    config = nose.config.Config()
    fpw_str = '--first-package-wins'
    opt_parser = config.getParser('')
    opt_def = opt_parser.get_option('--first-package-wins')
    if opt_def is None:
        raise RuntimeError('Nose does not accept "first-package-wins"'
                           ' - is this an old nose version?')
    if opt_def.takes_value(): # the =True variant
        fpw_str += '=True'
    return fpw_str


def prepare_imports():
    """ Prepare any imports for testing run

    At the moment, we prepare matplotlib by trying to make it use a backend that
    does not need a display
    """
    try:
        import matplotlib as mpl
    except ImportError:
        pass
    else:
        mpl.use('svg')


class NipyNoseTester(NoseTester):
    """ Numpy-like testing class

    * Removes some numpy-specific excludes
    * Disables numpy's fierce clearout of module import context for doctests
    * Run doctests by default

    """
    excludes = []

    def _get_custom_doctester(self):
        """ Use our our own doctester """
        import_nose()
        from .doctester import NipyDoctest
        return NipyDoctest()

    def test(self, label='fast', verbose=1, extra_argv=None, doctests=True,
             coverage=False):
        """
        Run tests for module using nose.

        As for numpy tester, except enable tests by default.

        Parameters
        ----------
        label : {'fast', 'full', '', attribute identifier}, optional
            Identifies the tests to run. This can be a string to pass to
            directly the nosetests executable with the '-A' option (an attribute
            identifier), or one of several special values.  Special values are:

            * 'fast' - the default - which corresponds to the ``nosetests -A``
              option of 'not slow'.
            * 'full' - fast (as above) and slow tests as in the
              'no -A' option to nosetests - this is the same as ''.
            * None or '' - run all tests.

        verbose : int, optional
            Verbosity value for test outputs, in the range 1-10. Default is 1.
        extra_argv : list, optional
            List with any extra arguments to pass to nosetests.
        doctests : bool, optional
            If True, run doctests in module. Default is True.
        coverage : bool, optional
            If True, report coverage of nipy code. Default is False.
            (This requires the `coverage module:
             <http://nedbatchelder.com/code/modules/coverage.html>`_).

        Returns
        -------
        result : object
            Returns the result of running the tests as a
            ``nose.result.TextTestResult`` object.

        Notes
        -----
        Each nipy module should expose `test` in its namespace to run all tests
        for it.  For example, to run all tests for nipy.algorithms:

        >>> import nipy.algorithms
        >>> nipy.algorithms.test() #doctest: +SKIP
        """
        prepare_imports()
        if extra_argv is None:
            extra_argv = []
        extra_argv.append(fpw_opt_str())
        return super(NipyNoseTester, self).test(label, verbose, extra_argv,
                                                doctests, coverage)



# -*- Mode: Python -*-  Not really, but the syntax is close enough
"""
Routines for massively univariate random-effect and mixed-effect
analysis. Two-sample case.

Author: Alexis Roche, 2008.
"""

__version__ = '0.1'


# Includes
from fff cimport *

# Exports from fff_twosample_stat.h
cdef extern from "fff_twosample_stat.h":

  ctypedef enum fff_twosample_stat_flag:
    FFF_TWOSAMPLE_STUDENT = 0,
    FFF_TWOSAMPLE_WILCOXON = 1,
    FFF_TWOSAMPLE_STUDENT_MFX = 10

  ctypedef struct fff_twosample_stat:
    pass

  ctypedef struct fff_twosample_stat_mfx:
    unsigned int niter

  fff_twosample_stat* fff_twosample_stat_new(unsigned int n1, unsigned int n2, fff_twosample_stat_flag flag)
  void fff_twosample_stat_delete(fff_twosample_stat* thisone) 
  double fff_twosample_stat_eval(fff_twosample_stat* thisone, fff_vector* x)

  fff_twosample_stat_mfx* fff_twosample_stat_mfx_new(unsigned int n1, unsigned int n2,
                                                     fff_twosample_stat_flag flag)
  void fff_twosample_stat_mfx_delete(fff_twosample_stat_mfx* thisone) 
  double fff_twosample_stat_mfx_eval(fff_twosample_stat_mfx* thisone, 
                                     fff_vector* x, fff_vector* vx)
  
  unsigned int fff_twosample_permutation(unsigned int* idx1, unsigned int* idx2, 
                                         unsigned int n1, unsigned int n2, double* magic)

  void fff_twosample_apply_permutation(fff_vector* px, fff_vector* pv, 
                                       fff_vector* x1, fff_vector* v1, 
                                       fff_vector* x2,  fff_vector* v2,
                                       unsigned int i, unsigned int* idx1, unsigned int* idx2)
   

# Initialize numpy
fffpy_import_array()
import_array()
import numpy as np


# Stat dictionary
stats = {'student': FFF_TWOSAMPLE_STUDENT,
         'wilcoxon': FFF_TWOSAMPLE_WILCOXON,
         'student_mfx': FFF_TWOSAMPLE_STUDENT_MFX}


def count_permutations(unsigned int n1, unsigned int n2):
  cdef double  n
  fff_twosample_permutation(NULL, NULL, n1, n2, &n)
  return int(n)
  

def stat(ndarray Y1, ndarray Y2, id='student', int axis=0, ndarray Magics=None):
  """
  T = stat(Y1, Y2, id='student', axis=0, magics=None).
  
  Compute a two-sample test statistic (Y1>Y2) over a number of
  deterministic or random permutations.
  """
  cdef fff_vector *y1, *y2, *t, *yp, *magics
  cdef fff_array *idx1, *idx2
  cdef unsigned int n, n1, n2, nex
  cdef unsigned long int simu, nsimu, idx
  cdef fff_twosample_stat* stat
  cdef fff_twosample_stat_flag flag_stat = stats[id]
  cdef double magic
  cdef fffpy_multi_iterator* multi

  # Get number of observations
  n1 = <unsigned int>Y1.shape[axis]
  n2 = <unsigned int>Y2.shape[axis]
  n = n1 + n2

  # Read out magic numbers
  if Magics == None:
    magics = fff_vector_new(1)
    magics.data[0] = 0 ## Just to make sure
  else:
    magics = fff_vector_fromPyArray(Magics)

  # Create output array
  nsimu = magics.size
  dims = [Y1.shape[i] for i in range(Y1.ndim)]
  dims[axis] = nsimu
  T = np.zeros(dims)

  # Create local structure
  yp = fff_vector_new(n)
  idx1 = fff_array_new1d(FFF_UINT, n1)
  idx2 = fff_array_new1d(FFF_UINT, n2)
  stat = fff_twosample_stat_new(n1, n2, flag_stat)

  # Multi-iterator 
  multi = fffpy_multi_iterator_new(3, axis, <void*>Y1, <void*>Y2, <void*>T)

  # Vector views
  y1 = multi.vector[0]
  y2 = multi.vector[1]
  t = multi.vector[2]
  
  # Loop
  for simu from 0 <= simu < nsimu:
    
    # Set the magic number
    magic = magics.data[simu*magics.stride]

    # Generate permutation 
    nex = fff_twosample_permutation(<unsigned int*>idx1.data,
                                    <unsigned int*>idx2.data,
                                    n1, n2, &magic)
   
    # Reset the multi-iterator
    fffpy_multi_iterator_reset(multi)
    
    # Perform the loop 
    idx = simu*t.stride
    while(multi.index < multi.size):
      fff_twosample_apply_permutation(yp, NULL, y1, NULL, y2, NULL, nex,
                                      <unsigned int*>idx1.data,
                                      <unsigned int*>idx2.data)
      t.data[idx] = fff_twosample_stat_eval(stat, yp)
      fffpy_multi_iterator_update(multi)

  # Delete local structures
  fffpy_multi_iterator_delete(multi)
  fff_vector_delete(magics)
  fff_vector_delete(yp)
  fff_array_delete(idx1)
  fff_array_delete(idx2)
  fff_twosample_stat_delete(stat)

  # Return
  return T


def stat_mfx(ndarray Y1, ndarray V1, ndarray Y2, ndarray V2,
             id='student_mfx', int axis=0, ndarray Magics=None,
             unsigned int niter=5):
  """
  T = stat(Y1, V1, Y2, V2, id='student', axis=0, magics=None, niter=5).
  
  Compute a two-sample test statistic (Y1>Y2) over a number of
  deterministic or random permutations.
  """
  cdef fff_vector *y1, *y2, *v1, *v2, *t, *yp, *vp, *magics
  cdef fff_array *idx1, *idx2
  cdef unsigned int n, n1, n2, nex
  cdef unsigned long int simu, nsimu, idx
  cdef fff_twosample_stat_mfx* stat
  cdef fff_twosample_stat_flag flag_stat = stats[id]
  cdef double magic
  cdef fffpy_multi_iterator* multi

  # Get number of observations
  n1 = <unsigned int>Y1.shape[axis]
  n2 = <unsigned int>Y2.shape[axis]
  n = n1 + n2

  # Read out magic numbers
  if Magics == None:
    magics = fff_vector_new(1)
    magics.data[0] = 0 ## Just to make sure
  else:
    magics = fff_vector_fromPyArray(Magics)

  # Create output array
  nsimu = magics.size
  dims = [Y1.shape[i] for i in range(Y1.ndim)]
  dims[axis] = nsimu 
  T = np.zeros(dims)

  # Create local structure
  yp = fff_vector_new(n)
  vp = fff_vector_new(n)
  idx1 = fff_array_new1d(FFF_UINT, n1)
  idx2 = fff_array_new1d(FFF_UINT, n2)
  stat = fff_twosample_stat_mfx_new(n1, n2, flag_stat)
  stat.niter = niter

  # Multi-iterator 
  multi = fffpy_multi_iterator_new(5, axis,
                                   <void*>Y1, <void*>V1,
                                   <void*>Y2, <void*>V2,
                                   <void*>T)

  # Vector views
  y1 = multi.vector[0]
  v1 = multi.vector[1]
  y2 = multi.vector[2]
  v2 = multi.vector[3]
  t = multi.vector[4]
  
  # Loop
  for simu from 0 <= simu < nsimu:
    
    # Set the magic number
    magic = magics.data[simu*magics.stride]

    # Generate permutation 
    nex = fff_twosample_permutation(<unsigned int*>idx1.data,
                                    <unsigned int*>idx2.data,
                                    n1, n2, &magic)
   
    # Reset the multi-iterator      
    fffpy_multi_iterator_reset(multi)
    
    # Perform the loop 
    idx = simu*t.stride
    while(multi.index < multi.size):
      fff_twosample_apply_permutation(yp, vp, y1, v1, y2, v2, nex,
                                      <unsigned int*>idx1.data,
                                      <unsigned int*>idx2.data)
      t.data[idx] = fff_twosample_stat_mfx_eval(stat, yp, vp)
      fffpy_multi_iterator_update(multi)

  # Delete local structures
  fffpy_multi_iterator_delete(multi)
  fff_vector_delete(magics)
  fff_vector_delete(yp)
  fff_vector_delete(vp)
  fff_array_delete(idx1)
  fff_array_delete(idx2)
  fff_twosample_stat_mfx_delete(stat)

  # Return
  return T



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#from _clustering import *
#from _clustering import __doc__

import numpy as np


def kmeans(X, nbclusters=2, Labels=None, maxiter=300, delta=0.0001, verbose=0,
              ninit=1):
    """ kmeans clustering algorithm

    Parameters
    ----------
    X: array of shape (n,p): n = number of items, p = dimension
       data array
    nbclusters (int), the number of desired clusters
    Labels = None array of shape (n) prior Labels.
           if None or inadequate a random initilization is performed.
    maxiter=300 (int), the maximum number of iterations  before convergence
    delta: float, optional,
           the relative increment in the results
           before declaring convergence.
    verbose: verbosity mode, optionall
    ninit: int, optional, number of random initalizations

    Returns
    -------
    Centers: array of shape (nbclusters, p),
             the centroids of  the resulting clusters
    Labels : array of size n, the discrete labels of the input items
    J (float):  the final value of the inertia criterion
    """
    nbitems = X.shape[0]
    if nbitems < 1:
        if verbose:
            raise ValueError(" I need at least one item to cluster")

    if np.size(X.shape) > 2:
        if verbose:
            raise ValueError("Please enter a two-dimensional array \
                              for clustering")

    if np.size(X.shape) == 1:
        X = np.reshape(X, (nbitems, 1))
    X = X.astype('d')

    nbclusters = int(nbclusters)
    if nbclusters < 1:
        if verbose:
            print " cannot compute less than 1 cluster"
        nbclusters = 1

    if nbclusters > nbitems:
        if verbose:
            print " cannot find more clusters than items"
        nbclusters = nbitems

    if (ninit < 1) & verbose:
        print "making at least one iteration"
        ninit = np.maximum(int(ninit), 1)

    if Labels != None:
        if np.size(Labels) == nbitems:
            Labels = Labels.astype(np.int)
            OK = (Labels.min() > -1) & (Labels.max() < nbclusters + 1)
            if OK:
                maxiter = int(maxiter)
                if maxiter > 0:
                    delta = float(delta)
                    if delta < 0:
                        if verbose:
                            print "incorrect stopping criterion - ignored"
                        delta = 0.0001
                    else:
                        pass
                else:
                    if verbose:
                        print "incorrect number of iterations - ignored"
                    maxiter = 300
            else:
                if verbose:
                    print "incorrect labelling - ignored"
        else:
            if verbose:
                print "incompatible number of labels provided - ignored"
    Centers, labels, J = _kmeans(X, nbclusters, Labels, maxiter, delta, ninit)
    return Centers, labels, J


def _MStep(x, z, k):
    """Computation of cluster centers/means

    Parameters
    ----------
    x array of shape (n,p)
      where n = number of samples, p = data dimension
    z, array of shape (x.shape[0]) current assignment
    k, int, number of desired clusters

    Returns
    -------
    centers, array of shape (k,p)
             the resulting centers
    """
    dim = x.shape[1]
    centers = np.repeat(np.reshape(x.mean(0), (1, dim)), k, 0)
    for q in range(k):
        if np.sum(z == q) == 0:
            pass
        else:
            centers[q] = np.mean(x[z == q], 0)
    return centers


def _EStep(x, centers):
    """ Computation of the input-to-cluster assignment

    Parameters
    ----------
    x array of shape (n,p)
      n = number of items, p = data dimension
    centers, array of shape (k,p) the cluster centers

    Returns
    -------
    z vector of shape(n), the resulting assignment
    """
    nbitem = x.shape[0]
    z = - np.ones(nbitem).astype(np.int)
    mindist = np.inf * np.ones(nbitem)
    k = centers.shape[0]
    for q in range(k):
        dist = np.sum((x - centers[q]) ** 2, 1)
        z[dist < mindist] = q
        mindist = np.minimum(dist, mindist)
    J = mindist.sum()
    return z, J


def voronoi(x, centers):
    """ Assignment of data items to nearest cluster center

    Parameters
    ----------
    x array of shape (n,p)
      n = number of items, p = data dimension
    centers, array of shape (k, p) the cluster centers

    Returns
    -------
    z vector of shape(n), the resulting assignment
    """
    if np.size(x) == x.shape[0]:
        x = np.reshape(x, (np.size(x), 1))
    if np.size(centers) == centers.shape[0]:
        centers = np.reshape(centers, (np.size(centers), 1))
    if x.shape[1] != centers.shape[1]:
        raise ValueError("Inconsistent dimensions for x and centers")

    return _EStep(x, centers)[0]


def _kmeans(X, nbclusters=2, Labels=None, maxiter=300, delta=1.e-4,
            ninit=1, verbose=0):
    """ kmeans clustering algorithm

    Parameters
    ----------
    X: array of shape (n,p): n = number of items, p = dimension
       data array
    nbclusters (int), the number of desired clusters
    Labels: array of shape (n) prior Labels, optional
            if None or inadequate a random initilization is performed.
    maxiter: int, optional
             the maximum number of iterations  before convergence
    delta: float, optional
           the relative increment in the results before declaring convergence.
    verbose=0: verboseity mode

    Returns
    -------
    Centers: array of shape (nbclusters, p),
             the centroids of  the resulting clusters
    Labels: array of size n, the discrete labels of the input items
    J, float,  the final value of the inertia criterion
    """
    # fixme: do the checks
    nbitem = X.shape[0]

    vdata = np.mean(np.var(X, 0))
    bJ = np.inf
    for it in range(ninit):
        # init
        if Labels == None:
            seeds = np.argsort(np.random.rand(nbitem))[:nbclusters]
            centers = X[seeds]
        else:
            centers = _MStep(X, Labels, nbclusters)
        centers_old = centers.copy()

        # iterations
        for i in range(maxiter):
            z, J = _EStep(X, centers)
            centers = _MStep(X, z, nbclusters)
            if verbose:
                print i, J
            if np.sum((centers_old - centers) ** 2) < delta * vdata:
                if verbose:
                    print i
                break
            centers_old = centers.copy()

            if J < bJ:
                bJ = J
                centers_output = centers.copy()
                z_output = z.copy()
    else:
        centers_output = centers
        z_output = z

    return centers_output, z_output, bJ



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import unittest

import numpy as np
import scipy.stats as st

from ..random_threshold import randthresh_main

def make_data(n=10, dim=20, r=5, mdim=15, maskdim=20, amplitude=10, 
                            noise=1, jitter=None, activation=False):
    XYZvol = np.zeros((dim,dim,dim),int)
    XYZ = np.array(np.where(XYZvol==0))
    p = XYZ.shape[1]
    #mask = np.arange(p)
    XYZvol[XYZ[0],XYZ[1],XYZ[2]] = np.arange(p)
    o = np.array([dim/2,dim/2,dim/2])
    I = XYZvol[(dim-mdim)/2:(dim+mdim)/2,(dim-mdim)/2:(dim+mdim)/2,(dim-mdim)/2:(dim+mdim)/2].ravel()
    mask = XYZvol[ (dim-maskdim)/2 : (dim+maskdim)/2, (dim-maskdim)/2 : (dim+maskdim)/2, (dim-maskdim)/2 : (dim+maskdim)/2 ].ravel()
    q = len(mask)
    maskvol = np.zeros((dim,dim,dim),int)
    maskvol[XYZ[0,mask],XYZ[1,mask],XYZ[2,mask]] = np.arange(q)
    Isignal = maskvol[dim/2-r:dim/2+r,dim/2-r:dim/2+r,dim/2-r:dim/2+r].ravel()
    signal = np.zeros(q,float)
    signal[Isignal] += amplitude
    X = np.zeros((n,p),float) + np.nan
    data = np.zeros((n,p),float) + np.nan
    vardata = np.zeros((n,p),float) + np.nan
    for i in xrange(n):
        X[i,I] = np.random.randn(len(I))
        if activation:
            o = np.array([dim/2,dim/2,dim/2])
            if jitter!=None:
                # numpy 2 casting rules do no allow in place addition of ints
                # and floats - hence not in-place here
                o = o + np.round(np.random.randn(3)*jitter).clip(r-mdim/2,mdim/2-r)
            #print o
        Ii = XYZvol[o[0]-r:o[0]+r,o[1]-r:o[1]+r,o[2]-r:o[2]+r].ravel()
        X[i,Ii] += amplitude
        vardata[i,I] = np.square(np.random.randn(len(I)))*noise**2
        data[i,I] = X[i,I] + np.random.randn(len(I))*np.sqrt(vardata[i,I])
    return data, XYZ, mask, XYZvol, vardata, signal

class TestRandomThreshold(unittest.TestCase):
    
    def test_random_threshold(self):
         # Just run all random threshold functions on toy data
         # for smoke testing
         data, XYZ, mask, XYZvol, vardata, signal = make_data(n=1, dim=20, r=3, mdim=20, maskdim=20, amplitude=4, noise=0, jitter=0, activation=True)
         Y = data[0]
         X = np.clip(-np.log(1 - st.chi2.cdf(Y**2, 1, 0)), 0, 1e10)
         K = (signal == 0).sum() - 100
         verbose=False
         randthresh_main(X, K, XYZ=None, p=np.inf, varwind=True, 
                         knownull=True, stop=False, verbose=verbose)
         randthresh_main(Y, K, XYZ=None, p=np.inf, varwind=True,
                         knownull=False, stop=True, verbose=verbose)
         
         randthresh_main(X, K, XYZ=None, p=np.inf, varwind=False, 
                         knownull=True, stop=True, verbose=verbose)
         randthresh_main(Y, K, XYZ=None, p=np.inf, varwind=False,
                         knownull=False, stop=False, verbose=verbose)
         
         randthresh_main(X, K, XYZ=XYZ, p=np.inf, varwind=True,
                         knownull=True, stop=False, verbose=verbose)
         randthresh_main(Y, K, XYZ=XYZ, p=np.inf, varwind=True,
                         knownull=False, stop=False, verbose=verbose)
         
         randthresh_main(X, K, XYZ=XYZ, p=np.inf, varwind=False,
                         knownull=True, stop=True, verbose=verbose)
         randthresh_main(Y, K, XYZ=XYZ, p=np.inf, varwind=False,
                         knownull=False, stop=True, verbose=verbose)


if __name__ == "__main__":
    unittest.main()

